import traceback
HAS_DEPS = True
try:
    from pandas import DataFrame
    import cv2
    import numpy as np
    import matplotlib.pyplot as plt
    import os
    import unittest
    import h5py
    import math
    import sys
    import arcgis
    import xml.etree.ElementTree as ET
    from datetime import datetime
    from arcgis.geometry import Polygon
    from IPython.display import clear_output
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_DEPS = False

def generate_kernel(k_size, k_type):
    """
        Generates kernel for morphological operations.
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        k_size                  List that determines the size of the kernel.

        ---------------------   -------------------------------------------
        k_type                  String that determines type of the kernel.
                                values can be "rect", "elliptical", "cross"
        =====================   ===========================================

        return:
        closing_kernel: Closing kernel object
        opening_kernel: Opening kernel object

     """
    closing_kernel, opening_kernel, kernel_object = None, None, None
    if k_type == "rect":
        kernel_object = cv2.MORPH_RECT
    if k_type == "elliptical":
        kernel_object = cv2.MORPH_ELLIPSE
    if k_type == "cross":
        kernel_object = cv2.MORPH_CROSS

    if k_size[0] is not None:
        closing_kernel = cv2.getStructuringElement(kernel_object, (k_size[0], k_size[0]))

    if k_size[1] is not None:
        opening_kernel = cv2.getStructuringElement(kernel_object, (k_size[1], k_size[1]))

    return closing_kernel, opening_kernel


def fill_mask_pixels(image, masked_path, land_color):
    """
        Fill the extracted mask region with the color of land region.
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        image                   Input image.

        ---------------------   -------------------------------------------
        masked_path             Path of extracted masked region image

        ---------------------   -------------------------------------------
        land_color              Land color to be filled at masked region

        =====================   ===========================================
        return:
        image: Image with colored masked replaced by land color
     """

    masked_image = cv2.imread(os.path.join(masked_path))
    mask = cv2.inRange(masked_image, np.array([0, 0, 0]), np.array([50, 50, 50]))
    h = image.shape[0]
    w = image.shape[1]
    for y in range(0, h):
        for x in range(0, w):
            if mask[y][x] == 0:
                image[y, x] = land_color

    return image


def detect_template(template, search_img, start_scale, end_scale, num_scales):
    """
        Performs multi-scale template matching to locate the template on the
        bigger search region
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        template                Template image which needs to be matched on
                                the bigger search region image

        ---------------------   -------------------------------------------
        search_img              Search region image in which template
                                needs to be searched

        ---------------------   -------------------------------------------
        start_scale             Starting value of scale interval

        ---------------------   -------------------------------------------
        end_scale               Ending value of scale interval

        ---------------------   -------------------------------------------
        num_scales              Number of scales within the scale interval


        =====================   ===========================================

        returns:
            start_x: Starting X co-ordinate of detected bounding box
            start_y: Starting Y co-ordinate of detected bounding box
            end_x: Ending X co-ordinate of detected bounding box
            end_y: Ending Y co-ordinate of detected bounding box
            max_val: Value of template match score for the detected bounding box
            scale_factor: Scale at which the match was found for the template
    """

    template_height = template.shape[0]
    template_width = template.shape[1]

    unittest.TestCase().assertGreaterEqual(search_img.shape[0], template_height,
                                           '''Height of the search image should be \
                                           greater then the height of template image''')
    unittest.TestCase().assertGreaterEqual(search_img.shape[1], template_width,
                                           '''Width of the search image should be \
                                           greater then the width of template image''')

    start_x = 0
    start_y = 0
    end_x = 0
    end_y = 0
    scale_factor = -1
    found = None
    for scale in np.linspace(start_scale, end_scale, num_scales)[::-1]:
        resize_width, resize_height = search_img.shape[1] * scale, \
                                      search_img.shape[0] * scale
        resized = cv2.resize(search_img,
                             (int(resize_width),
                              int(resize_height)))
        scale_factor = search_img.shape[1] / float(resized.shape[1])
        if resized.shape[0] < template_height or resized.shape[1] < template_width:
            break
        result = cv2.matchTemplate(resized, template, cv2.TM_CCOEFF_NORMED)
        (_, max_val, _, max_loc) = cv2.minMaxLoc(result)
        if found is None or max_val > found[0]:
            found = (max_val, max_loc, scale_factor)

    if found is not None:
        (max_val, max_loc, scale_factor) = found
        (start_x, start_y) = (int(max_loc[0] * scale_factor),
                              int(max_loc[1] * scale_factor))
        (end_x, end_y) = (int((max_loc[0] + template_width) * scale_factor),
                          int((max_loc[1] + template_height) * scale_factor))
    else:
        max_val = -1

    return start_x, start_y, end_x, end_y, max_val, scale_factor


def compute_reference_homography(bb_start_x, bb_start_y,
                                 bb_end_x, bb_end_y,
                                 search_region_height, search_region_width,
                                 template_height, template_width):
    """
    Computes the homography transform using matched location of scanned map on
    search region
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        bb_start_x              Bounding box left x co-ordinate

        ---------------------   -------------------------------------------
        bb_start_y              Bounding box top y co-ordinate

        ---------------------   -------------------------------------------
        bb_end_x                Bounding box right x co-ordinate

        ---------------------   -------------------------------------------
        template_height         Height of template image

        ---------------------   -------------------------------------------
        template_width          Width of template image


        =====================   ===========================================

    returns:
        homography_matrix : Computed transformation matrix
    """

    src_points = np.array([[0, 0],
                           [template_width - 1, 0],
                           [template_width - 1, template_height - 1],
                           [0, template_height - 1]],
                          dtype=np.float32)

    dst_points = np.array([[bb_start_x, bb_start_y],
                           [bb_end_x, bb_start_y],
                           [bb_end_x, bb_end_y],
                           [bb_start_x, bb_end_y]],
                          dtype=np.float32)
    homography_matrix = cv2.getPerspectiveTransform(src_points, dst_points)

    return homography_matrix


def apply_homography(query_img, search_region, homography_matrix):
    """
        Creates warped scanned map on search region using estimated homography
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        query_img               Image to be processed

        ---------------------   -------------------------------------------
        search_region           Image for search region

        ---------------------   -------------------------------------------
        homography_matrix       Transformed matrix to be used
        =====================   ===========================================

        returns:
            warped_query_img: Modified world search region image
    """

    search_region_height = search_region.shape[0]
    search_region_width = search_region.shape[1]

    warped_query_img = cv2.warpPerspective(query_img, homography_matrix,
                                           (search_region_width, search_region_height))

    return warped_query_img


def check_bb_boundary(bb_start_x, bb_start_y,
                      bb_end_x, bb_end_y,
                      height, width):
    """
    Checks the boundary and modifies the values if necessary
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        bb_start_x              X co-ordinate of left top corner of bounding
                                box
        ---------------------   -------------------------------------------
        bb_start_y              Y co-ordinate of left top corner of bounding
                                box
        ---------------------   -------------------------------------------
        bb_end_x                X co-ordinate of right bottom corner of
                                bounding box
        ---------------------   -------------------------------------------
        bb_end_y                Y co-ordinate of right bottom corner of
                                bounding box
        ---------------------   -------------------------------------------
        height                  Height of the image

        ---------------------   -------------------------------------------
        width                   Width of the image

        =====================   ===========================================

    returns:
        bb_start_x: Modified X co-ordinate of left top corner of bounding box
        bb_start_y: Modified Y co-ordinate of left top corner of bounding box
        bb_end_x: Modified X co-ordinate of right bottom corner of bounding box
        bb_end_y: Modified Y co-ordinate of right bottom corner of bounding box
    """

    if bb_start_x < 0:
        bb_start_x = 0

    if bb_start_y < 0:
        bb_start_y = 0

    if bb_end_x >= width:
        bb_end_x = width - 1

    if bb_end_y >= height:
        bb_end_y = height - 1
    return bb_start_x, bb_start_y, bb_end_x, bb_end_y


def detect_contours(input_img):
    """
        Detects contours in the input image
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        input_img               Input image for which contours need to be
                                computed

        =====================   ===========================================

        returns:
            contours: List of computed contours for the input image
    """

    if input_img.ndim != 2:
        input_img_gray = cv2.cvtColor(input_img, cv2.COLOR_BGR2GRAY)
    else:
        input_img_gray = input_img

    ret, input_img_gray_thresh = cv2.threshold(input_img_gray, 127,
                                               255, cv2.THRESH_BINARY)

    contours, hierarchy = cv2.findContours(input_img_gray_thresh, cv2.RETR_EXTERNAL,
                                           cv2.CHAIN_APPROX_SIMPLE)

    return contours


def calculate_diagonal_length(length, height):
    """
        Find the length of diagonal of a rectangle
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        length                  Length of rectangle
        ---------------------   -------------------------------------------
        height                  Height of rectangle
        =====================   ===========================================

        returns:
            length_diag: Length of the diagonal
    """

    length_diag = math.sqrt(length ** 2 + height ** 2)

    return length_diag


def calculate_nonzero_ratio(input_img):
    """
        Calculates the ratio of non-zero pixels area w.r.t entire area of image
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        input_img               Image for which ratio needs to be computed

        =====================   ===========================================

        returns:
            ratio: Non-zero area ratio
    """

    non_zero_area = cv2.countNonZero(input_img)
    total_area = input_img.shape[0] * input_img.shape[1]

    ratio = non_zero_area / total_area

    return ratio


def calculate_distance_ratios(bb_start_x, bb_start_y,
                              bb_end_x, bb_end_y,
                              point_x, point_y):
    """
        Calculates the distance ratio w.r.t X and Y co-ordinate of a point
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        bb_start_x              X co-ordinate of left top corner of bounding
                                box
        ---------------------   -------------------------------------------
        bb_start_y              Y co-ordinate of left top corner of bounding
                                box
        ---------------------   -------------------------------------------
        bb_end_x                X co-ordinate of right bottom corner of
                                bounding box
        ---------------------   -------------------------------------------
        bb_end_y                Y co-ordinate of right bottom corner of
                                bounding box
        ---------------------   -------------------------------------------
        point_x                 X co-ordinate of point

        ---------------------   -------------------------------------------
        point_y                 Y co-ordinate of point
        =====================   ===========================================

        returns:
            ratio_x : X co-ordinate ratio
            ratio_y : Y co-ordinate ratio
    """

    epsilon = 0.00001

    distance_point_x_start_x = point_x - bb_start_x
    distance_point_x_end_x = bb_end_x - point_x

    distance_point_y_start_y = point_y - bb_start_y
    distance_point_y_end_y = bb_end_y - point_y

    ratio_x = (distance_point_x_start_x
               / (distance_point_x_end_x + epsilon))

    ratio_y = (distance_point_y_start_y
               / (distance_point_y_end_y + epsilon))

    return ratio_x, ratio_y


def find_distance(point1_x, point1_y,
                  point2_x, point2_y):
    """
    Find distance between two points
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        point1_x                X co-ordinate of first point
        ---------------------   -------------------------------------------
        point1_y                Y co-ordinate of first point
        ---------------------   -------------------------------------------
        point2_x                X co-ordinate of second point
        ---------------------   -------------------------------------------
        point2_y                y co-ordinate of second point

        =====================   ===========================================

    return:
        distance: distance between two points
    """

    distance = math.sqrt((point1_x - point2_x) ** 2 +
                         (point1_y - point2_y) ** 2)
    return distance


def find_nearest_contour_point(input_contour, pt_x, pt_y):
    """
    Find point on contour which is nearest to the given point
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        input_contour            Contour on which points need to be
                                 evaluated for nearness
        ---------------------   -------------------------------------------
        pt_x                     X co-ordinate of point
        ---------------------   -------------------------------------------
        pt_y                     Y co-ordinate of point

        =====================   ===========================================

    returns:
        min_x: X co-ordinate of contour point nearest to input point
        min_y: Y co-ordinate of contour point nearest to input point
        min_index: Index of contour point nearest to input point
        min_distance: Distance of contour point nearest to input point
    """
    min_distance = float("inf")
    min_index = -1
    min_x = -1
    min_y = -1

    for i, contour_pt in enumerate(input_contour):
        distance = find_distance(contour_pt[0][0], contour_pt[0][1], pt_x, pt_y)

        if distance < min_distance:
            min_distance = distance
            min_index = i
            min_x = contour_pt[0][0]
            min_y = contour_pt[0][1]

    return min_x, min_y, min_index, min_distance


def create_random_color_list(size):
    """
    Generate a list of specified size with random color values
        =====================   ==================================
        **Argument**            **Description**
        ---------------------   ----------------------------------
         size                    Size of the list which needs to
                                 be returned
        =====================   ==================================

    returns:
       color_list: List of random color values
    """

    color_list = []

    for i in range(0, size):
        color_tuple = tuple(np.random.choice(range(256), size=3))
        color_tuple = tuple([int(x) for x in color_tuple])
        color_list.append(color_tuple)

    return color_list


def annotate_control_pts(query_image, search_region, control_pts):
    """
    Annotate control points on search region and scanned map
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        query_image             Input image
        ---------------------   -------------------------------------------
        search_region           Color image of search region
        ---------------------   -------------------------------------------
        control_pts             Control points to be annotated

        =====================   ===========================================

    returns:
        search_region_img_updated: Annotated search region image
        query_image_updated: Annotated scanned map
    """
    search_region_img_copy = search_region.copy()
    query_image_copy = query_image.copy()
    colors = create_random_color_list(len(control_pts))

    for i, control_pt in enumerate(control_pts):
        color = colors[i]
        color = tuple([int(x) for x in color])

        query_image_copy = cv2.drawMarker(query_image_copy,
                                          (control_pt[0], -1 * control_pt[1]), color)

        search_region_img_copy = cv2.drawMarker(search_region_img_copy,
                                                (int(control_pt[5]), int(-1 * control_pt[6])), color)

    return search_region_img_copy, query_image_copy


def compute_refined_homography_outlier_mask(control_pts):
    """
    Computes refined homography and outlier mask by applying RANSAC
     on mapped control points
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        control_pts             List of control points in both scanned map
                                and search region

        =====================   ===========================================

    returns:
        homography_matrix: Computed transformed matrix
        outlier_mask: Mask array with flags indicating whether control points
                      are outliers (denoted by 0) or inliers(denoted by 1)
    """

    if len(control_pts) < 4:
        return None, None

    src_points = np.zeros((len(control_pts), 2), dtype=np.float32)
    dst_points = np.zeros((len(control_pts), 2), dtype=np.float32)

    for i, control_pt in enumerate(control_pts):
        src_points[i, 0] = control_pt[0]
        src_points[i, 1] = -1 * control_pt[1]
        dst_points[i, 0] = control_pt[5]
        dst_points[i, 1] = -1 * control_pt[6]

    homography_matrix, outlier_mask = cv2.findHomography(src_points, dst_points,
                                                         cv2.RANSAC, 3)

    return homography_matrix, outlier_mask


def filter_control_pts(control_pts, outlier_mask):
    """
    Filter outliers from control points using outlier mask
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        control_pts             List of control points in scanned map and
                                corresponding control points in search
                                region image
        ---------------------   -------------------------------------------
        outlier_mask            Outlier mask with flags indicating whether
                                outlier or not

        =====================   ===========================================

    returns:
        filtered_control_pts: Control Points without outliers
    """
    filtered_control_pts = []

    for i, control_pt in enumerate(control_pts):
        if outlier_mask[i] == 1:
            filtered_control_pts.append(control_pt)

    return filtered_control_pts


def estimate_control_points(search_reg_roi, template,
                            search_reg_roi_offset_x, search_reg_roi_offset_y,
                            search_reg_roi_start_x, search_reg_roi_start_y,
                            search_region_img, extent):
    """
    Detects contours in template image and finds correspondences
    in processed search region image ROI

        =========================   ====================================
        **Argument**                **Description**
        -------------------------   -----------------------------------
        search_reg_roi               Transformed search region image ROI
        -------------------------    -----------------------------------
        template                     Template image
        -------------------------    ------------------------------------
        search_reg_roi_offset_x      X margin for search region ROI w.r.t
                                     warped template
        -------------------------    ------------------------------------
        search_reg_roi_offset_y      Y margin for search region ROI w.r.t
                                     warped template
        -------------------------    -------------------------------------
        search_reg_roi_start_x       X co-ordinate of left top corner of
                                     search region roi
        -------------------------    -------------------------------------
        search_reg_roi_offset_y      Y co-ordinate of left top corner of
                                     search region roi
        -------------------------    --------------------------------------
        search_region_img            Transformed search region image ROI
        -------------------------    --------------------------------------
        extent                       Search region extent
        -------------------------    --------------------------------------

        =========================    ===========================

    returns:
        mapped_control_pts: List of control points in template and
        corresponding points in search image
    """

    degree_per_pixel_x = (extent["xmax"] - extent["xmin"]) \
                         / search_region_img.shape[1]

    degree_per_pixel_y = (extent["ymin"] - extent["ymax"]) \
                         / search_region_img.shape[0]

    template_contours = detect_contours(template)
    biggest_template_contour = max(template_contours, key=cv2.contourArea)

    contour_pts = []

    bb_cnt_xmin, bb_cnt_ymin, \
    bb_cnt_width, bb_cnt_height = cv2.boundingRect(biggest_template_contour)

    bb_cnt_area = bb_cnt_width * bb_cnt_height
    template_area = template.shape[0] * template.shape[1]

    area_ratio = bb_cnt_area / template_area

    if area_ratio < 0.7:
        for contour in template_contours:
            for contour_pt in contour:
                contour_pts.append(contour_pt)

    else:
        contour_pts = biggest_template_contour

    length_diagonal_template_img = calculate_diagonal_length(template.shape[1],
                                                             template.shape[0])
    contour_patch_win = int(length_diagonal_template_img
                            / 8)

    delta_contour_patch_win = contour_patch_win / 2

    search_reg_patch_win = int(contour_patch_win * 1.5)
    delta_search_patch_win = (search_reg_patch_win / 2)

    mapped_control_pts = []

    for contour_pt in contour_pts:
        contour_pt_x = contour_pt[0][0]
        contour_pt_y = contour_pt[0][1]
        cont_patch_start_y = contour_pt_y - delta_contour_patch_win
        cont_patch_end_y = contour_pt_y + delta_contour_patch_win
        cont_patch_start_x = contour_pt_x - delta_contour_patch_win
        cont_patch_end_x = contour_pt_x + delta_contour_patch_win

        cont_patch_start_x, cont_patch_start_y, \
        cont_patch_end_x, cont_patch_end_y = check_bb_boundary(cont_patch_start_x,
                                                               cont_patch_start_y,
                                                               cont_patch_end_x,
                                                               cont_patch_end_y,
                                                               template.shape[0],
                                                               template.shape[1])

        cont_patch = template[int(cont_patch_start_y): int(cont_patch_end_y), \
                     int(cont_patch_start_x): int(cont_patch_end_x)]

        search_reg_start_y = contour_pt_y - delta_search_patch_win + search_reg_roi_offset_y
        search_reg_end_y = contour_pt_y + delta_search_patch_win + search_reg_roi_offset_y
        search_reg_start_x = contour_pt_x - delta_search_patch_win + search_reg_roi_offset_x
        search_reg_end_x = contour_pt_x + delta_search_patch_win + search_reg_roi_offset_x

        (search_reg_start_x, search_reg_start_y,
         search_reg_end_x, search_reg_end_y) = check_bb_boundary(search_reg_start_x, search_reg_start_y,
                                                                 search_reg_end_x, search_reg_end_y,
                                                                 search_reg_roi.shape[0], search_reg_roi.shape[1])

        search_reg = search_reg_roi[int(search_reg_start_y):int(search_reg_end_y),
                     int(search_reg_start_x):int(search_reg_end_x)]

        (detection_start_x, detection_start_y,
         detection_end_x, detection_end_y,
         detection_val, scale_factor) = detect_template(cont_patch, search_reg,
                                                        1.0, 1.0,
                                                        1)
        search_reg_contours = detect_contours(search_reg)

        if detection_val > 0:
            ratio_x, ratio_y = calculate_distance_ratios(cont_patch_start_x,
                                                         cont_patch_start_y,
                                                         cont_patch_end_x,
                                                         cont_patch_end_y,
                                                         contour_pt_x,
                                                         contour_pt_y)

            search_reg_centroid_x = ((ratio_x * detection_end_x)
                                     + detection_start_x) / (ratio_x + 1)
            search_reg_centroid_y = ((ratio_y * detection_end_y) +
                                     detection_start_y) / (ratio_y + 1)

            search_reg_contour_min_x = float("inf")
            search_reg_contour_min_y = float("inf")
            search_reg_min_distance = float("inf")

            for search_reg_contour in search_reg_contours:
                (min_x, min_y,
                 min_index, min_distance) = find_nearest_contour_point(
                    search_reg_contour,
                    search_reg_centroid_x,
                    search_reg_centroid_y)

                if min_distance < search_reg_min_distance:
                    search_reg_min_distance = min_distance
                    search_reg_contour_min_x = min_x
                    search_reg_contour_min_y = min_y

            search_reg_centroid_x = search_reg_contour_min_x
            search_reg_centroid_y = search_reg_contour_min_y

            centroid_search_reg_detection_x = search_reg_centroid_x + search_reg_start_x + \
                                              search_reg_roi_start_x
            centroid_search_reg_detection_y = search_reg_centroid_y + search_reg_start_y + \
                                              search_reg_roi_start_y

            centroid_search_reg_detection_x_long = extent["xmin"] + (degree_per_pixel_x *
                                                                     centroid_search_reg_detection_x)
            centroid_search_reg_detection_y_lat = extent["ymax"] + (degree_per_pixel_y *
                                                                    centroid_search_reg_detection_y)

            mapped_control_pts.append([contour_pt_x, -1 * contour_pt_y,
                                       centroid_search_reg_detection_x_long,
                                       centroid_search_reg_detection_y_lat,
                                       detection_val, centroid_search_reg_detection_x,
                                       -1 * centroid_search_reg_detection_y])

    return mapped_control_pts


def calculate_iou_binary_images(image1, image2):
    """
    Calculate iou using two binary images
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        image1                  First binary image
        ---------------------   -------------------------------------------
        image2                  Second binary image

        =====================   ===========================================

    returns:
        iou: Computed IOU value
    """
    intersection_img = cv2.bitwise_and(image1, image2)
    intersection_pixels = cv2.countNonZero(intersection_img)

    union_img = cv2.bitwise_or(image1, image2)
    union_pixels = cv2.countNonZero(union_img)

    iou = intersection_pixels / union_pixels
    return iou


def compare_homography_matrices(reference_img, refined_img, template, control_pts):
    """
    Compares and selects the best homography transform between reference homography
    and refined homography
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        reference_img           Image for computed reference homography
        ---------------------   -------------------------------------------
        refined_img             Image for computed refined homography
        ---------------------   -------------------------------------------
        template                Template image
        ---------------------   -------------------------------------------
        control_pts             Control points

        =====================   ===========================================

    returns:
        index: Index of selected transformed matrix
    """

    control_pts_to_use = control_pts
    src_points = np.zeros((len(control_pts_to_use), 2), dtype=np.float32)
    min_x = float("inf")
    max_x = -1
    min_y = float("inf")
    max_y = -1

    for i, control_pt in enumerate(control_pts_to_use):
        src_points[i, 0] = control_pt[0]
        src_points[i, 1] = -1 * control_pt[1]

        if src_points[i, 0] > max_x:
            max_x = src_points[i, 0]
        if src_points[i, 0] < min_x:
            min_x = src_points[i, 0]

        if src_points[i, 1] > max_y:
            max_y = src_points[i, 1]
        if src_points[i, 1] < min_y:
            min_y = src_points[i, 1]

    area = (max_x - min_x) * (max_y - min_y)
    area_non_zero = cv2.countNonZero(template)

    area_ratio = area / area_non_zero

    reference_img_gray = cv2.cvtColor(reference_img, cv2.COLOR_BGR2GRAY)
    refined_img_gray = cv2.cvtColor(refined_img, cv2.COLOR_BGR2GRAY)

    ret, reference_img_mask = cv2.threshold(reference_img_gray, 0, 255, cv2.THRESH_BINARY)
    ret, refined_img_mask = cv2.threshold(refined_img_gray, 0, 255, cv2.THRESH_BINARY)

    iou = calculate_iou_binary_images(reference_img_mask,
                                      refined_img_mask)

    if area_ratio > 0.35 and iou > 0.70:
        return 1
    else:
        return 0


def map_species_region(region_mask_img, search_region,
                       reference_homography, refined_homography,
                       selected_index, detection_start_x,
                       detection_start_y, detection_end_x,
                       detection_end_y, color, output_path,
                       image_name, extent, idx, region):
    """
        Maps the species region masks on the search region using the computed
        homography matrices
        =====================   ===================================================
        **Argument**            **Description**
        ---------------------   ----------------------------------------------------
        region_mask_img          Species region mask depicting species distribution
        ---------------------   ----------------------------------------------------
        search_region            Color image of search region
        ---------------------   ----------------------------------------------------
        reference_homography     Computed reference homography
        ---------------------   ----------------------------------------------------
        refined_homography       Computed refined homography
        ---------------------   ----------------------------------------------------
        selected_index           Index indicating which transformed matrix is the
                                 final computed homography
        ---------------------   ----------------------------------------------------
        detection_start_x        X co-ordinate of the top left corner of detected
                                 bounding box
        ---------------------   ----------------------------------------------------
        detection_start_y        Y co-ordinate of the top left corner of detected
                                 bounding box
        ---------------------   ----------------------------------------------------
        detection_end_x          X co-ordinate of the bottom right corner of
                                 detected bounding box
        ---------------------   ----------------------------------------------------
        detection_end_y          Y co-ordinate of the bottom right corner of
                                 detected bounding box
        ---------------------   ----------------------------------------------------
        color                    Color of the current masked region to be mapped
        ---------------------   ----------------------------------------------------
        output_path              Path to generate shapefile output
        ---------------------   ----------------------------------------------------
        image_name               Name of the image
        ---------------------   ----------------------------------------------------
        extent                   Selected extent for search region
        ---------------------   ----------------------------------------------------
        idx                      Index of the current scanned map in the folder
        ---------------------   ----------------------------------------------------
        region                   Index of the current masked region

        =====================   ====================================================

        returns:
            mapped_species_region: Search region with species region mapped
    """

    search_region_height = search_region.shape[0]
    search_region_width = search_region.shape[1]

    warped_img = cv2.warpPerspective(region_mask_img, reference_homography,
                                     (search_region_width, search_region_height))

    if selected_index == 1:
        img_to_warp = warped_img[detection_start_y: detection_end_y, \
                      detection_start_x: detection_end_x]
        warped_img = cv2.warpPerspective(img_to_warp, refined_homography,
                                         (search_region_width, search_region_height))

    warped_img_contours = detect_contours(warped_img)
    data = write_shapefile(warped_img_contours, output_path,
                           image_name, extent, search_region_height,
                           search_region_width, idx,
                           region)

    modified_search_region = search_region.copy()
    modified_search_region = cv2.drawContours(modified_search_region,
                                              warped_img_contours,
                                              -1, (int(color[2]),
                                                   int(color[1]),
                                                   int(color[0])),
                                              cv2.FILLED)

    return modified_search_region, data


def process_contour(input_contour):
    """
         Adds end points to contours such that the endpoint is same as
         the start point and reverses the direction of
         the contour
        =====================   ================================
        **Argument**            **Description**
        ---------------------   --------------------------------
        input_contour            contour to be processed

        =====================   ================================

    returns:
        final_contour: Processed contour
    """

    if len(input_contour) == 0:
        return

    processed_contour = np.flip(input_contour, 0)
    final_contour = []
    for point in processed_contour:
        final_contour.append(point)

    final_contour.append(processed_contour[0])
    return final_contour


def find_current_timestamp():
    """
        Creates a string representing the current timestamp

        :returns:
            timestamp_str: Current timestamp string in the format "%H_%M_%S_%f"
    """
    datetime_obj = datetime.now()
    timestamp_str = datetime_obj.strftime("%Y_%m_%d_%H_%M_%S_%f")

    return timestamp_str


def combine_shapefiles(output_dir, file_name, data, extent):
    """
        Combines shapefile outputs in a particular folder
        =====================   ==================================================
        **Argument**            **Description**
        ---------------------   --------------------------------------------------
        output_dir               Output directory where combined shapefile needs
                                 to be dumped
        ---------------------   --------------------------------------------------
        file_name                Filename of the combined shapefile
        ---------------------   --------------------------------------------------
        data                     Data to generate the shapefile
        ---------------------   --------------------------------------------------
        extent                   Selected extent for search region

        =====================   ==================================================

        returns:
            None
    """
    data_frame = DataFrame(data, columns=['FID', 'SHAPE', 'SPECIES', 'SCANNED'])

    data_frame['SHAPE'] = data_frame['SHAPE'].apply(Polygon)
    data_frame.spatial.set_geometry('SHAPE', extent["spatialReference"]["wkid"])

    file_path = os.path.join(output_dir, file_name + ".shp")

    data_frame.spatial.to_featureclass(file_path)


def write_shapefile(input_contours, output_dir,
                    file_name, extent,
                    search_region_height, search_region_width,
                    idx, region):
    """
        Creates data frame using input contours for feature layer creation
        =====================   ===================================================
        **Argument**            **Description**
        ---------------------   ----------------------------------------------------
        input_contours           Contours which need to be transformed to lat-long
                                 co-ordinates
        ---------------------   ----------------------------------------------------
        output_dir               Output directory where combined shapefile needs
                                 to be dumped
        ---------------------   ----------------------------------------------------
        file_name                Filename of the combined shapefile
        ---------------------   ----------------------------------------------------
        extent                   Selected extent for search region
        ---------------------   ----------------------------------------------------
        search_region_height     Height of search region image
        ---------------------   ----------------------------------------------------
        search_region_width      Width of search region image
        ---------------------   ----------------------------------------------------
        idx                      Index of the current scanned map in the folder
        ---------------------   ----------------------------------------------------
        region                   Index of the current masked region

        =====================   ====================================================

        returns:
            None
    """

    if len(input_contours) == 0:
        return
    shape = dict()
    shape['rings'] = []
    data_temp = []

    for index, contour in enumerate(input_contours):
        contour = process_contour(contour)
        list_long_lat = calculate_contour_lat_long(contour,
                                                   extent,
                                                   search_region_height,
                                                   search_region_width)
        shape['rings'].append(list_long_lat)

    shape['spatialReference'] = {'wkid': extent["spatialReference"]["wkid"]}
    shape["hasZ"] = False
    shape["hasM"] = False

    species_name = file_name
    scanned_img = file_name + ".jpg"
    data_temp.append([idx, shape, species_name, scanned_img])

    data_frame = DataFrame(data_temp, columns=['FID', 'SHAPE', 'SPECIES', 'SCANNED'])

    data_frame['SHAPE'] = data_frame['SHAPE'].apply(Polygon)
    data_frame.spatial.set_geometry('SHAPE', extent["spatialReference"]["wkid"])

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    if not os.path.exists(os.path.join(output_dir, "region" + str(region))):
        os.makedirs(os.path.join(output_dir, "region" + str(region)))

    file_path = os.path.join(output_dir, "region" + str(region), file_name + ".shp")

    data_frame.spatial.to_featureclass(file_path)
    return [idx, shape, species_name, scanned_img]


def calculate_contour_lat_long(input_contour, extent,
                               search_region_height, search_region_width):
    """
        Calculates latitude and longitude co-ordinates for x and y image
        co-ordinates of contour points
        =====================   =============================================
        **Argument**            **Description**
        ---------------------   ---------------------------------------------
        input_contours           Contours which need to be transformed
                                 to lat-long co-ordinates
        ---------------------   ---------------------------------------------
        extent                   Selected extent for search region
        ---------------------   ---------------------------------------------
        search_region_height     Height of search region image
        ---------------------   ---------------------------------------------
        search_region_width      Width of search region image

        =====================   =============================================

        returns:
            list_long_lat: List of latitude and longitude co-ordinates
             corresponding to x and y image co-ordinates of contour points

    """
    degree_per_pixel_x = (extent["xmax"] - extent["xmin"]) / search_region_width

    degree_per_pixel_y = (extent["ymin"] - extent["ymax"]) / search_region_height
    list_long_lat = []

    for point in input_contour:
        long_val = extent["xmin"] + (degree_per_pixel_x * point[0][0])
        lat_val = extent["ymax"] + (degree_per_pixel_y * point[0][1])
        long_lat = [long_val, lat_val]
        list_long_lat.append(long_lat)

    return list_long_lat


def write_georeference_xml_file(control_point, path, image_name, extent):
    """
        Writes XML file used to geo-reference the image on the search image
        =====================   ===================================================
        **Argument**            **Description**
        ---------------------   ----------------------------------------------------
        control_point            Calculated control points for corresponding scanned
                                 map
        ---------------------   ----------------------------------------------------
        path                     Path to dump XML file
        ---------------------   ----------------------------------------------------
        image_name               Name of scanned map

        =====================   ====================================================

        returns:
            None
    """

    cps = control_point
    aux_template = """
    <PAMDataset>
      <Metadata domain="IMAGE_STRUCTURE">
        <MDI key="INTERLEAVE">PIXEL</MDI>
      </Metadata>
      <Metadata domain="xml:ESRI" format="xml">
        <GeodataXform xsi:type="typens:PolynomialXform" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xs="
        http://www.w3.org/2001/XMLSchema" xmlns:typens="http://www.esri.com/schemas/ArcGIS/2.5.0">
          <PolynomialOrder>1</PolynomialOrder>
          <SpatialReference xsi:type="typens:GeographicCoordinateSystem">
            <WKT>GEOGCS["GCS_WGS_1984",DATUM["D_WGS_1984",SPHEROID["WGS_1984",6378137.0,298.257223563]],PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433],AUTHORITY["EPSG",4326]]</WKT>
            <XOrigin>-400</XOrigin>
            <YOrigin>-400</YOrigin>
            <XYScale>999999999.99999988</XYScale>
            <ZOrigin>-100000</ZOrigin>
            <ZScale>10000</ZScale>
            <MOrigin>-100000</MOrigin>
            <MScale>10000</MScale>
            <XYTolerance>8.983152841195215e-09</XYTolerance>
            <ZTolerance>0.001</ZTolerance>
            <MTolerance>0.001</MTolerance>
            <HighPrecision>true</HighPrecision>
            <LeftLongitude>""" + str(extent["xmin"]) + """</LeftLongitude>
            <WKID>""" + str(extent["spatialReference"]["wkid"]) + """</WKID>
            <LatestWKID>""" + str(extent["spatialReference"]["wkid"]) + """</LatestWKID>
          </SpatialReference>
          <SourceGCPs xsi:type="typens:ArrayOfDouble">
            <Double>201.77370325693613</Double>
            <Double>-74.731001206272538</Double>
            <Double>389.95419492620601</Double>
            <Double>-1362.5314876417463</Double>
            <Double>1216.3773578347073</Double>
            <Double>-428.69359007958542</Double>
            <Double>101.70061599187169</Double>
            <Double>-594.21366542213968</Double>
          </SourceGCPs>
          <TargetGCPs xsi:type="typens:ArrayOfDouble">
            <Double>72.481240538209477</Double>
            <Double>35.906754227992636</Double>
            <Double>77.413842048833715</Double>
            <Double>8.1204837096300242</Double>
            <Double>97.145971270897093</Double>
            <Double>27.129739103324169</Double>
            <Double>70.554967027409319</Double>
            <Double>24.491757967748939</Double>
          </TargetGCPs>
          <Name />
        </GeodataXform>
      </Metadata>
    </PAMDataset>
    """.strip()

    aux = ET.ElementTree(ET.fromstring(aux_template))
    gf = aux.findall('Metadata')[1].find('GeodataXform')
    source_gcps = gf.find('SourceGCPs')
    target_gcps = gf.find('TargetGCPs')

    for child in source_gcps.findall('Double'):
        source_gcps.remove(child)

    for child in target_gcps.findall('Double'):
        target_gcps.remove(child)

    for cp in cps:
        image_x = ET.SubElement(source_gcps, "Double")
        image_x.text = str(cp[0])
        image_y = ET.SubElement(source_gcps, "Double")
        image_y.text = str(cp[1])
        map_x = ET.SubElement(target_gcps, "Double")
        map_x.text = str(cp[2])
        map_y = ET.SubElement(target_gcps, "Double")
        map_y.text = str(cp[3])

    gf.set('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
    gf.set('xmlns:xs', 'http://www.w3.org/2001/XMLSchema')
    gf.set('xmlns:typens', 'http://www.esri.com/schemas/ArcGIS/2.5.0')

    aux_file_out = os.path.join(path, os.path.basename(image_name) + '.jpg.aux.xml')
    aux.write(aux_file_out, encoding='utf-8', )


def convert_to_xy(lat_val, long_val, extent,
                  degree_per_pixel_x, degree_per_pixel_y):
    """
    Converts latitude and longitude to image x and y co-ordinates
        =====================   ====================================
        **Argument**            **Description**
        ---------------------   ------------------------------------
        lat_val                  Latitude value of the search
                                 region image point
        ---------------------   ------------------------------------
        long_val                 Longitude value of the search
                                 region image point
        ---------------------   ------------------------------------
        degree_per_pixel_x       Calculated degree w.r.t X axis
        ---------------------   ------------------------------------
        degree_per_pixel_y       Calculated degree w.r.t Y axis

        =====================   ====================================

    returns:
        [x_val, y_val]: Tuple with image x and y co-ordinates corresponding to
        input latitude and longitude co-ordinates
    """
    extreme_left = extent["xmin"]
    extreme_top = extent["ymax"]

    x_val = (long_val - extreme_left) / degree_per_pixel_x

    y_val = (lat_val - extreme_top) / degree_per_pixel_y
    return [int(x_val), int(y_val)]


def calculate_contour_xy(list_lat_long, extent,
                         degree_per_pixel_x, degree_per_pixel_y):
    """
        Calculates image x and y co-ordinates for all points corresponding to
        latitude and longitude co-ordinates
        =====================   =======================================
        **Argument**            **Description**
        ---------------------   ---------------------------------------
        list_lat_long            List of all lat-long pairs
                                 corresponding to points
        ---------------------   ---------------------------------------
        extent                   Selected extent for search region
        ---------------------   ---------------------------------------
        degree_per_pixel_x       Calculated degree w.r.t X axis
        ---------------------   ---------------------------------------
        degree_per_pixel_y       Calculated degree w.r.t Y axis
        =====================   =======================================

    returns:
        contour_xy: List of x-y pairs corresponding to points
    """
    contour_xy = []
    for point in list_lat_long:
        xy = convert_to_xy(point[1], point[0], extent, degree_per_pixel_x, degree_per_pixel_y)
        contour_xy.append([xy])

    return contour_xy


def make_world_map_from_shapefile(image_path, image_height,
                                  image_width, extent, degree_per_pixel_x,
                                  degree_per_pixel_y, color):
    """
        Creates search image using shapefile
        =====================   =========================================
        **Argument**            **Description**
        ---------------------   -----------------------------------------
        image_path               Path of the scanned map
        ---------------------   -----------------------------------------
        image_height             Height of scanned map
        ---------------------   -----------------------------------------
        image_width              Height of scanned map
        ---------------------   -----------------------------------------
        extent                   Selected extent for search region
        ---------------------   -----------------------------------------
        degree_per_pixel_x       Calculated degree w.r.t X axis
        ---------------------   -----------------------------------------
        degree_per_pixel_y       Calculated degree w.r.t Y axis
        ---------------------   -----------------------------------------
        color                    r, g, b value for water color
        =====================   =========================================

    returns:
        world_map_img: Color image of the search image
    """

    data_frame = arcgis.features.GeoAccessor.from_featureclass(os.path.join(image_path))

    contours = []
    max_length = -1
    for index in data_frame.index:
        contour = calculate_contour_xy(data_frame['SHAPE'][index]['rings'][0], extent, degree_per_pixel_x,
                                       degree_per_pixel_y)
        length = len(contour)
        if length > max_length:
            max_length = length

        contours.append(np.array(contour))

    world_map_img = np.zeros((image_height, image_width, 3), dtype=np.uint8)
    world_map_img[:, :, 0] = color[2]
    world_map_img[:, :, 1] = color[1]
    world_map_img[:, :, 2] = color[0]
    world_map_img = cv2.drawContours(world_map_img, contours,
                                     -1, (255, 255, 255), cv2.FILLED)

    return world_map_img


def generate_search_template(search_image, process_folder, x1, y1, x2, y2):
    """
        Crops search template from the bigger image using a given extent and converts to
        binary image
        =====================   ==========================================
        **Argument**            **Description**
        ---------------------   ------------------------------------------
        search_image             Search Image
        ---------------------   ------------------------------------------
        process_folder           Path of the intermediate result folder
        ---------------------   ------------------------------------------
        x1                       Start index of column
        ---------------------   ------------------------------------------
        y1                       Start index of row
        ---------------------   ------------------------------------------
        x2                       End index of column
        ---------------------   ------------------------------------------
        y2                       End index of row
        =====================   ==========================================

    returns:
        transformed_img: transformed image of the search region
    """
    all_images = os.listdir(os.path.join(process_folder))
    display_img = search_image
    image = search_image
    image = np.zeros(search_image.shape, np.uint8)
    image[y1:y2, x1:x2] = search_image[y1:y2, x1:x2]

    if image.ndim != 2:
        image_gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    else:
        image_gray = image

    transformed_img = np.zeros((image_gray.shape[0],
                                image_gray.shape[1]), dtype=np.uint8)

    image_height = image_gray.shape[0]
    image_width = image_gray.shape[1]

    for y in range(0, image_height):
        for x in range(0, image_width):
            if image_gray[y][x] >= 245:
                transformed_img[y][x] = 255
            else:
                transformed_img[y][x] = 0

    for image_path in all_images:
        path = os.path.join(process_folder, image_path, "search_region")
        if not os.path.isdir(path):
            os.mkdir(path)
        cv2.imwrite(os.path.join(path, "search_region_rgb.jpg"), display_img)
        cv2.imwrite(os.path.join(path, "search_region_template.jpg"), transformed_img)

    return transformed_img


def generate_search_coordinate(extent, bigger_extreme_left,
                               bigger_extreme_top, bigger_extreme_right,
                               bigger_extreme_bottom, degree_per_pixel_x,
                               degree_per_pixel_y):
    """
        Generates co-ordinate for search region
        =====================   ============================================
        **Argument**            **Description**
        ---------------------   --------------------------------------------
        extent                   Selected extent for search region
        ---------------------   --------------------------------------------
        bigger_extreme_left      Extreme left co-ordinate of world image
        ---------------------   --------------------------------------------
        bigger_extreme_top       Extreme top co-ordinate of world image
        ---------------------   --------------------------------------------
        bigger_extreme_right     Extreme right co-ordinate of world image
        ---------------------   --------------------------------------------
        bigger_extreme_bottom    Extreme bottom co-ordinate of world image
        ---------------------   --------------------------------------------
        degree_per_pixel_x       Calculated degree w.r.t X axis
        ---------------------   --------------------------------------------
        degree_per_pixel_y       Calculated degree w.r.t Y axis

        =====================   ============================================

    returns:
    x1, y1, x2, y2 to crop extent from search region
    """

    extreme_left = extent["xmin"]
    extreme_top = extent["ymax"]
    extreme_right = extent["xmax"]
    extreme_bottom = extent["ymin"]

    x1 = int(((abs(bigger_extreme_left) - abs(extreme_left)) / abs(degree_per_pixel_x)))
    y1 = int(((abs(bigger_extreme_top) - abs(extreme_top)) / abs(degree_per_pixel_y)))
    y2 = int(((abs(bigger_extreme_bottom) + abs(extreme_bottom)) / abs(degree_per_pixel_y)))
    x2 = int(((abs(bigger_extreme_right) + abs(extreme_right)) / abs(degree_per_pixel_x)))
    return x1, y1, x2, y2


def plot_image(*images):
    """
        Displays output
        =====================   ============================================
        **Argument**            **Description**
        ---------------------   --------------------------------------------
        images                   List of images and their descriptions
        =====================   ============================================
    returns:
        None
    """
    all_images = images[0]
    all_names = images[1]
    size = images[2]
    fig, ax = plt.subplots(1, len(all_images), figsize=(size, size))

    for i, img in enumerate(all_images):
        ax[i].axis('off')
        font_dict = {
            'fontsize': 10
        }
        ax[i].set_aspect(aspect=0.5)
        ax[i].title.set_position([.5, -0.07 * len(all_images)])
        ax[i].set_title(all_names[i], fontdict=font_dict)
        ax[i].imshow(img)


def display_progress_bar(i, max_images, postText):
    """
        Displays progress bar
        =====================   =======================================
        **Argument**            **Description**
        ---------------------   ---------------------------------------
        i                        Index of the image inside the loop
        ---------------------   ---------------------------------------
        max_images               Maximum number of images
        ---------------------   ---------------------------------------
        postText                 Text to be added at the end of the
                                 progress bar
        =====================   =======================================
    returns:
        None
    """
    n_bar = 50
    j = i / max_images
    sys.stdout.write('\r Status: ')
    sys.stdout.write(f"[{'=' * int(n_bar * j):{n_bar}s}] {int(100 * j)}%  {postText}")
    sys.stdout.flush()


class ScannedMapDigitizer:
    """
    Creates the object for ScannedMapDigitizer class

    =====================   ============================================
    **Argument**            **Description**
    ---------------------   --------------------------------------------
    input_folder             Path to the folder that contains extracted
                             maps
    ---------------------   --------------------------------------------
    output_folder            Path to the folder where intermediate
                             results should get generated

    =====================   ============================================

    """
    input_path = None
    process_path = None
    all_images_path = []
    extent_data = {}

    def __init__(self, input_folder, output_folder):
        if not HAS_DEPS:
            print("**Environment fails**")
            raise Exception(
                f"""{import_exception} \n\nThis module requires opencv, pandas,
                                        numpy and matplotlib as its dependencies.""")
        else:
            self.initialize_variables(input_folder, output_folder)

    @classmethod
    def get_search_region_extent(cls):
        """
        Getter function for search region extent
        """
        try:
            extent = {
                'spatialReference': {'wkid': cls.wkid},
                'xmin': cls.logitude_extreme_left,
                'ymax': cls.latitude_extreme_top,
                'xmax': cls.logitude_extreme_right,
                'ymin': cls.latitude_extreme_bottom
            }
        except AttributeError:
            print("Search region extent is not set, please use set_search_region_extent() to set the search extent. ")
            extent = {}
        return extent

    @classmethod
    def set_search_region_extent(cls, extent):

        """
        Creates the object for ScannedMapDigitizer class

        =====================   ============================================
        **Argument**            **Description**
        ---------------------   --------------------------------------------
        extent                  Extent defines the extreme longitude/latitude
                                of the search region.

        =====================   ============================================

        """

        cls.wkid = extent["spatialReference"]["wkid"]
        cls.logitude_extreme_left = extent["xmin"]
        cls.latitude_extreme_top = extent["ymax"]
        cls.logitude_extreme_right = extent["xmax"]
        cls.latitude_extreme_bottom = extent["ymin"]
        return "Search region extent updated successfully !!"

    @classmethod
    def initialize_variables(cls, input_folder, output_folder):
        cls.input_path = input_folder
        cls.process_path = output_folder
        cls.all_images_path = []

    @classmethod
    def create_mask(cls, color_list, color_delta=60,
                    kernel_size=None, kernel_type="rect",
                    show_result=True):
        """
        Generates the binary masked images

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        color_list              A list containing different color inputs
                                (r, g, b).
        ---------------------   -------------------------------------------
        color_delta             A value which defines the range around the
                                threshold value for a specific color used
                                for creating the mask images.
                                Default value is 60
        ---------------------   -------------------------------------------
        kernel_size             A list of 2 integers corresponding to size
                                of the morphological filter operations
                                closing and opening respectively.
        ---------------------   -------------------------------------------
        kernel_type             A string value defining the type/shape of
                                the kernel. kernel type can be "rect",
                                "elliptical" or "cross".
                                Default value is "rect".
        ---------------------   -------------------------------------------
        show_result             A boolean value. Set to "True" to visualize
                                results and set to "False" otherwise
        =====================   ===========================================
        """

        input_folder = cls.input_path
        output_folder = cls.process_path
        cls.all_images_path = []
        all_images = os.listdir(os.path.join(input_folder))

        if not os.path.isdir(output_folder):
            os.mkdir(output_folder)

        if kernel_size is not None:
            unittest.TestCase().assertEqual(len(kernel_size), 2, '''Expected length of \
                                                                    kernel_size is 2''')
            closing, opening = generate_kernel(kernel_size, kernel_type)
        else:
            closing, opening = [None, None]

        for idx, image_path in enumerate(all_images):
            image_name = image_path.split(".")[0]
            cls.all_images_path.append(image_name)
            image_folder = os.path.join(output_folder, image_name)
            if not os.path.isdir(image_folder):
                os.mkdir(image_folder)
            image_folder = os.path.join(image_folder, "mask")
            if not os.path.isdir(image_folder):
                os.mkdir(image_folder)

            image = cv2.imread(os.path.join(input_folder, image_path))
            rgb_img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            hf = h5py.File(os.path.join(output_folder, image_name,
                                        "mask", "mask_color_details.h5"), 'w')
            img_masks = [rgb_img]
            img_name = ["Scanned Map"]
            for i, color in enumerate(color_list):
                clr = np.array([[color]])
                mask = cv2.inRange(rgb_img, clr - color_delta, clr + color_delta)
                background = np.full(rgb_img.shape, 255, dtype=np.uint8)
                image_output = cv2.bitwise_or(background, background, mask=mask)
                if closing is not None:
                    image_output = cv2.morphologyEx(image_output, cv2.MORPH_CLOSE, closing)
                if opening is not None:
                    image_output = cv2.morphologyEx(image_output, cv2.MORPH_OPEN, opening)

                cv2.imwrite(os.path.join(output_folder, image_name, "input.jpg"), image)
                cv2.imwrite(os.path.join(image_folder, "region_mask_" + str(i) + ".jpg"),
                            image_output)
                hf.create_dataset("region_mask_" + str(i) + ".jpg", data=color)
                img_masks.append(image_output)
                img_name.append("Region Binary Mask")

            hf.close()
            clear_output(wait=True)
            display_progress_bar(idx + 1, len(all_images), "Completed \n\n\r")
            if show_result:
                plot_image(img_masks, img_name, 10)

        print("Output generated at ", output_folder)

    @classmethod
    def create_template_image(cls, color,
                              color_delta=10, kernel_size=2,
                              show_result=True):
        """
       This method generates templates and color masks from scanned maps which
       are used in the subsequent step of template matching.

        =====================   ===============================================
        **Argument**            **Description**
        ---------------------   -----------------------------------------------
        color                   r, g, b value representing land color.
                                The color parameter is required for extracting
                                the land region and generating the binary mask
        ---------------------   -----------------------------------------------
        color_delta             A value which defines the range around the
                                threshold value for a specific color used for
                                creating the mask images.
                                 Default value is 60
        ---------------------   -----------------------------------------------
        kernel_size             A list of 2 integers corresponding to size of
                                the morphological filter operations closing
                                and opening respectively.
        ---------------------   -----------------------------------------------
        show_result             A Boolean value. Set to "True" to visualize
                                results and set to "False" otherwise
        =====================   ===============================================

        """
        process_folder = cls.process_path
        all_images = cls.all_images_path
        clr = np.array(color)
        for idx, image_path in enumerate(all_images):
            image = cv2.imread(os.path.join(process_folder, image_path, "input.jpg"))
            rgb_img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            display_images = [cv2.cvtColor(image, cv2.COLOR_BGR2RGB)]
            image_name = ["Scanned Map"]
            all_masked_images = os.listdir(os.path.join(process_folder, image_path, "mask"))
            all_masked_images = [img_name for img_name in all_masked_images \
                                 if "region_mask_" in img_name]
            for masked_image in all_masked_images:
                masked_path = os.path.join(process_folder, image_path, "mask", masked_image)
                mask_filled_image = fill_mask_pixels(rgb_img, masked_path, color)
                if kernel_size is not None:
                    kernel = np.ones((kernel_size, kernel_size), np.uint8)
                    rgb_img = cv2.dilate(mask_filled_image, kernel, iterations=1)
                else:
                    rgb_img = mask_filled_image

            black_mask = cv2.inRange(rgb_img, clr - color_delta, clr + color_delta)
            background = np.full(rgb_img.shape, 255, dtype=np.uint8)
            image_output = cv2.bitwise_or(background, background, mask=black_mask)
            template_image_path = os.path.join(process_folder, image_path, "template_image")
            if not os.path.isdir(template_image_path):
                os.mkdir(template_image_path)
            cv2.imwrite(os.path.join(template_image_path, "masked_template.jpg"), image_output)
            display_images.append(image_output)
            image_name.append("Binary Template")
            clear_output(wait=True)
            display_progress_bar(idx + 1, len(all_images), "Completed \n\n\r")
            if show_result:
                plot_image(display_images, image_name, 12)

        print("Output generated at:", process_folder)

    @classmethod
    def prepare_search_region(cls, search_image,
                              color, extent,
                              image_height, image_width,
                              show_result=True):
        """
        This method prepares the search region in which the prepared templates are
        to be searched.

        =====================   ==================================================
        **Argument**            **Description**
        ---------------------   --------------------------------------------------
        color                   r, g, b value representing water color
        ---------------------   --------------------------------------------------
        search_image            Path to the bigger image/shapefile.
        ---------------------   --------------------------------------------------
        extent                  Extent defines the extreme longitude/latitude
                                of the search region.
        ---------------------   --------------------------------------------------
        image_height            Height of the search region.
        ---------------------   --------------------------------------------------
        image_width             Width of the search region.
        ---------------------   --------------------------------------------------
        show_result             A boolean value. Set to "True" to visualize
                                results and set to "False" otherwise
        =====================   ==================================================
        """

        cls.extent_data = extent
        try:
            bigger_extreme_left = cls.logitude_extreme_left
            bigger_extreme_top = cls.latitude_extreme_top
            bigger_extreme_right = cls.logitude_extreme_right
            bigger_extreme_bottom = cls.latitude_extreme_bottom
        except AttributeError:
            cls.set_search_region_extent(extent)
            bigger_extreme_left = cls.logitude_extreme_left
            bigger_extreme_top = cls.latitude_extreme_top
            bigger_extreme_right = cls.logitude_extreme_right
            bigger_extreme_bottom = cls.latitude_extreme_bottom

        bigger_extent = {
            "xmin": bigger_extreme_left,
            "ymax": bigger_extreme_top,
            "xmax": bigger_extreme_right,
            "ymin": bigger_extreme_bottom
        }

        degree_per_pixel_x = (bigger_extreme_right - bigger_extreme_left) / image_width
        degree_per_pixel_y = (bigger_extreme_bottom - bigger_extreme_top) / image_height

        if search_image.lower().endswith(('.png', '.jpg',
                                          '.jpeg', '.tiff',
                                          '.tif', '.bmp',
                                          '.gif')):
            image = cv2.imread(os.path.join(search_image))
            image = cv2.resize(image, (image_width, image_height))

        elif search_image.lower().endswith('.shp'):
            print("Processing Shape file...")
            image = make_world_map_from_shapefile(search_image, image_height,
                                                  image_width, bigger_extent,
                                                  degree_per_pixel_x, degree_per_pixel_y,
                                                  color)
            print("Done.")
        else:
            raise NameError('Invalid input image or file extention!!')

        print("Extracting given extent from the world imagery...")
        process_folder = cls.process_path
        x1, y1, x2, y2 = generate_search_coordinate(extent, bigger_extreme_left,
                                                    bigger_extreme_top, bigger_extreme_right,
                                                    bigger_extreme_bottom, degree_per_pixel_x,
                                                    degree_per_pixel_y)
        processed_image = generate_search_template(image, process_folder,
                                                   x1, y1,
                                                   x2, y2)

        path = os.path.join(process_folder, "search_region")
        if not os.path.isdir(path):
            os.mkdir(path)
        print("Done.")
        cv2.imwrite(os.path.join(path, "search_template.jpg"),
                    processed_image)
        cv2.imwrite(os.path.join(path, "search_rgb.jpg"),
                    image)
        title = ["Search Region", "Binary Search Region"]
        if show_result:
            plot_image([cv2.cvtColor(image, cv2.COLOR_BGR2RGB),
                        cv2.cvtColor(processed_image, cv2.COLOR_GRAY2RGB)],
                       title, 13)

    @classmethod
    def match_template_multiscale(cls, min_scale,
                                  max_scale, num_scales,
                                  show_result=True):
        """
        This method finds the location of the best match of a smaller image
        (template) in a larger image(search image) assuming it exists in the
        larger image.

        =====================   ============================================
        **Argument**            **Description**
        ---------------------   --------------------------------------------
        min_scale               An integer representing the minimum scale
                                at which template matching is performed
        ---------------------   --------------------------------------------
        max_scale               An integer representing maximum scale at
                                which template matching is performed
        ---------------------   --------------------------------------------
        num_scales              An integer representing the number
                                of scales at which template matching is
                                performed.
        ---------------------   --------------------------------------------
        show_result             A Boolean value. Set to "True" to visualize
                                results and set to "False" otherwise
        =====================   ============================================

        """
        process_folder = cls.process_path

        all_images = cls.all_images_path

        region_name = []
        region_accuracy = []

        for idx, image_path in enumerate(all_images):
            image = cv2.imread(os.path.join(process_folder, image_path, "input.jpg"))
            display_images = [cv2.cvtColor(image, cv2.COLOR_BGR2RGB)]
            image_template = cv2.imread(
                os.path.join(process_folder, image_path,
                             "template_image", "masked_template.jpg"), 0)
            search_image = cv2.imread(
                os.path.join(process_folder, image_path,
                             "search_region", "search_region_rgb.jpg"))
            search_image_template = cv2.imread(
                os.path.join(process_folder, image_path,
                             "search_region", "search_region_template.jpg"), 0)
            detection_start_x, detection_start_y, \
            detection_end_x, detection_end_y, \
            max_val, scale_factor = detect_template(image_template, search_image_template,
                                                    min_scale, max_scale,
                                                    num_scales)

            region_name.append(image_path)
            region_accuracy.append(round(max_val, 2))
            reference_homography = compute_reference_homography(detection_start_x,
                                                                detection_start_y,
                                                                detection_end_x,
                                                                detection_end_y,
                                                                search_image_template.shape[0],
                                                                search_image_template.shape[1],
                                                                image_template.shape[0],
                                                                image_template.shape[1]
                                                                )
            path = os.path.join(process_folder, image_path, "template_match")
            if not os.path.isdir(path):
                os.mkdir(path)

            hf = h5py.File(os.path.join(path, "reference_homography.h5"), 'w')
            hf.create_dataset('detection_points',
                              data=(detection_start_x, detection_start_y,
                                    detection_end_x, detection_end_y))
            hf.create_dataset('reference_homography', data=reference_homography)
            hf.close()
            reference_image = apply_homography(image, search_image, reference_homography)

            cv2.imwrite(os.path.join(path, "reference_image.jpg"), reference_image)

            reference_world_image = cv2.addWeighted(search_image,
                                                    0.5,
                                                    reference_image,
                                                    0.5, 0)
            cv2.imwrite(os.path.join(path, "reference_world_image.jpg"),
                        reference_world_image)
            display_images.append(cv2.cvtColor(reference_world_image,
                                               cv2.COLOR_BGR2RGB))
            title = ["Scanned Map", "Match Accuracy: " + str(round(max_val, 3) * 100) + " %"]
            clear_output(wait=True)
            display_progress_bar(idx + 1, len(all_images), "Completed \n\n\r")
            if show_result:
                plot_image(display_images, title, 12)

        print("Output generated at: ", process_folder)

    @classmethod
    def georeference_image(cls, padding_param, show_result=True):

        """
        This method estimates the control point pairs by traversing the
        contours of template image and finding the corresponding matches
        on the search region ROI image

        =====================   ========================================
        **Argument**            **Description**
        ---------------------   ----------------------------------------
        padding_param           A tuple that contains x-padding
                                and y-padding at 0th and 1st index
                                respectively
        ---------------------   ----------------------------------------
        show_result             A Boolean value. Set to "True" to
                                visualize results and set to "False"
                                otherwise
        =====================   ========================================
        """

        process_folder = cls.process_path

        all_images = cls.all_images_path
        extent = cls.get_search_region_extent()

        for idx, image_path in enumerate(all_images):
            image = cv2.imread(os.path.join(process_folder, image_path, "input.jpg"))
            image_template = cv2.imread(
                os.path.join(process_folder, image_path,
                             "template_image",
                             "masked_template.jpg"), 0)
            search_image = cv2.imread(
                os.path.join(process_folder, image_path,
                             "search_region",
                             "search_region_rgb.jpg"))
            search_image_template = cv2.imread(
                os.path.join(process_folder, image_path,
                             "search_region",
                             "search_region_template.jpg"), 0)
            reference_image = cv2.imread(
                os.path.join(process_folder, image_path,
                             "template_match", "reference_image.jpg"))

            hf = h5py.File(os.path.join(process_folder,
                                        image_path,
                                        "template_match",
                                        "reference_homography.h5"), 'r')
            detection = hf.get('detection_points')
            (detection_start_x, detection_start_y,
             detection_end_x, detection_end_y) = np.array(detection)
            homography = hf.get('reference_homography')
            reference_homography = np.array(homography)
            hf.close()

            search_image_height = search_image.shape[0]
            search_image_width = search_image.shape[1]
            search_image_roi_margin_x = padding_param[0]
            search_image_roi_margin_y = padding_param[1]

            warped_processed_image = cv2.warpPerspective(image_template, reference_homography,
                                                         (search_image_width, search_image_height))
            warped_original_image = cv2.warpPerspective(image, reference_homography,
                                                        (search_image_width, search_image_height))

            warped_image = warped_original_image[detection_start_y: detection_end_y, \
                           detection_start_x: detection_end_x]
            warped_template = warped_processed_image[detection_start_y: detection_end_y, \
                              detection_start_x: detection_end_x]

            search_image_roi_start_x = detection_start_x - search_image_roi_margin_x
            search_image_roi_end_x = detection_end_x + search_image_roi_margin_x
            search_image_roi_start_y = detection_start_y - search_image_roi_margin_y
            search_image_roi_end_y = detection_end_y + search_image_roi_margin_y

            (search_image_roi_start_x,
             search_image_roi_start_y,
             search_image_roi_end_x,
             search_image_roi_end_y) = check_bb_boundary(
                search_image_roi_start_x,
                search_image_roi_start_y,
                search_image_roi_end_x,
                search_image_roi_end_y,
                search_image_template.shape[0],
                search_image_template.shape[1])

            search_image_offset_start_x = detection_start_x - search_image_roi_start_x
            search_image_offset_start_y = detection_start_y - search_image_roi_start_y

            search_image_roi_processed = search_image_template[
                                         int(search_image_roi_start_y): \
                                         int(search_image_roi_end_y),
                                         int(search_image_roi_start_x): \
                                         int(search_image_roi_end_x)]

            control_pts = estimate_control_points(search_image_roi_processed, warped_template,
                                                  search_image_offset_start_x, search_image_offset_start_y,
                                                  search_image_roi_start_x, search_image_roi_start_y,
                                                  search_image, extent)

            path = os.path.join(process_folder, image_path, "georeference_image")

            refined_homography, outlier_mask = \
                compute_refined_homography_outlier_mask(control_pts)

            filtered_control_pts = filter_control_pts(control_pts, outlier_mask)
            search_image_filtered_control_pts, \
            input_image_filtered_control_pts = annotate_control_pts(warped_image, search_image,
                                                                    filtered_control_pts)

            refined_image = apply_homography(warped_image, search_image,
                                             refined_homography)
            refined_search_image = cv2.addWeighted(search_image, 0.5,
                                                   refined_image, 0.5, 0)
            selected_index = compare_homography_matrices(reference_image, refined_image,
                                                         warped_template, control_pts)

            if not os.path.isdir(path):
                os.mkdir(path)
            write_georeference_xml_file(filtered_control_pts, path, image_path, extent)
            hf = h5py.File(os.path.join(path, "refined_homography.h5"), 'w')
            hf.create_dataset('refined_homography', data=refined_homography)
            hf.create_dataset('selected_index', data=[selected_index])
            hf.close()
            cv2.imwrite(os.path.join(path, "filtered_control_pts.jpg"),
                        input_image_filtered_control_pts)
            cv2.imwrite(os.path.join(path, "warped_image.jpg"),
                        warped_image)
            cv2.imwrite(os.path.join(path, image_path + ".jpg"),
                        warped_image)
            title = ["Control Points on Scanned Map", "Control Points on Search Region", \
                     "Transformed Image"]
            clear_output(wait=True)
            display_progress_bar(idx + 1, len(all_images), "Completed \n\n\r")
            if show_result:
                plot_image([cv2.cvtColor(input_image_filtered_control_pts,
                                         cv2.COLOR_BGR2RGB),
                            cv2.cvtColor(search_image_filtered_control_pts,
                                         cv2.COLOR_BGR2RGB),
                            cv2.cvtColor(refined_search_image,
                                         cv2.COLOR_BGR2RGB)], title, 15)

        print("Output generated at: ", process_folder)

    @classmethod
    def digitize_image(cls, show_result=True):

        """
        This method is the final step in the pipeline that maps the
        species regions on the search image using the computed
        transformations.
        Also, it generates the shapefiles for the species region that can be
        visualized using ArcGIS Pro and further edited.

        =====================   =============================================
        **Argument**            **Description**
        ---------------------   ---------------------------------------------
        show_result             A Boolean value. Set to "True" to visualize
                                results and set to "False" otherwise
        =====================   =============================================

        """
        process_folder = cls.process_path
        extent = cls.get_search_region_extent()

        all_images = cls.all_images_path
        all_combined = {}
        for idx, image_path in enumerate(all_images):
            data = []
            color_counter = 0
            image = cv2.imread(os.path.join(process_folder, image_path,
                                            "input.jpg"))
            search_image = cv2.imread(
                os.path.join(process_folder, image_path,
                             "search_region", "search_region_rgb.jpg"))

            hf = h5py.File(os.path.join(process_folder, image_path,
                                        "template_match", "reference_homography.h5"), 'r')
            detection = hf.get('detection_points')
            (detection_start_x, detection_start_y,
             detection_end_x, detection_end_y) = np.array(detection)
            homography = hf.get('reference_homography')
            reference_homography = np.array(homography)
            hf.close()

            hf = h5py.File(os.path.join(process_folder, image_path,
                                        "georeference_image", "refined_homography.h5"), 'r')
            homography = hf.get('refined_homography')
            refined_homography = np.array(homography)
            index = hf.get("selected_index")
            selected_index = int(np.array(index)[0])
            hf.close()

            path = os.path.join(process_folder, image_path, "digitize_map")
            if not os.path.isdir(path):
                os.mkdir(path)

            mapped_species_region_img = search_image
            hf = h5py.File(os.path.join(process_folder, image_path,
                                        "mask", "mask_color_details.h5"), 'r')
            all_masked_images = os.listdir(os.path.join(process_folder, image_path,
                                                        "mask"))
            all_masked_images = [img_name for img_name in all_masked_images \
                                 if "region_mask_" in img_name]

            for region, masked_image in enumerate(all_masked_images):
                color_counter += 1
                masked_path = os.path.join(process_folder, image_path,
                                           "mask", masked_image)
                region_mask_img = cv2.imread(masked_path)
                color = np.array(hf.get(masked_image), dtype="int64")
                (mapped_species_region_img,
                 shape_file_details) = map_species_region(region_mask_img, mapped_species_region_img,
                                                          reference_homography, refined_homography,
                                                          selected_index, detection_start_x,
                                                          detection_start_y, detection_end_x,
                                                          detection_end_y, color,
                                                          path, image_path,
                                                          extent, idx,
                                                          region)
                data.append(shape_file_details)
                if idx == 0:
                    all_combined[str(region)] = [shape_file_details]
                else:
                    all_combined[str(region)].append(shape_file_details)

            hf.close()

            cv2.imwrite(os.path.join(path, "map_with_specie_region.jpg"),
                        mapped_species_region_img)
            title = ["Scanned Map", "Transformed Region Mask"]
            clear_output(wait=True)
            display_progress_bar(idx + 1, len(all_images), "Completed \n\n\r")
            if show_result:
                plot_image([cv2.cvtColor(image, cv2.COLOR_BGR2RGB),
                            cv2.cvtColor(mapped_species_region_img,
                                         cv2.COLOR_BGR2RGB)], title, 12)

        combined_folder = os.path.join(process_folder, "combined_shapefiles")
        if not os.path.isdir(combined_folder):
            os.mkdir(combined_folder)
        for key, val in all_combined.items():
            if not os.path.isdir(os.path.join(combined_folder, "combined_region_" + str(key))):
                os.mkdir(os.path.join(combined_folder, "combined_region_" + str(key)))
            combine_shapefiles(os.path.join(combined_folder, "combined_region_" + str(key)),
                               "combined_region", val, extent)



from pathlib import Path
import json
from ._arcgis_model import _EmptyData, _change_tail, ArcGISModel
from ._codetemplate import code, image_classifier_prf
import warnings
import arcgis
import sys, os, importlib
from functools import partial
import logging
logger = logging.getLogger()

HAS_OPENCV = True
HAS_FASTAI = True
HAS_ARCPY = True

try:
    import torch
    from torch import nn
    import numpy as np
    from fastai.basic_train import Learner, LearnerCallback
    from fastai.torch_core import split_model_idx
    from fastai.vision import ImageList
    from fastai.vision import imagenet_stats, normalize
    from fastai.core import has_arg, split_kwargs_by_func
    from fastai.basic_data import DatasetType
    from fastai.callback import Callback
    from fastai.torch_core import to_cpu, grab_idx
    from fastai.basic_train import loss_batch
    from fastai.vision.image import open_image, bb2hw, image2np, Image, pil2tensor
    import PIL
    from ._ssd_utils import compute_class_AP
    from .._utils.pascal_voc_rectangles import show_results_multispectral, ObjectDetectionCategoryList
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from ._arcgis_model import _set_ddp_multigpu, _isnotebook
    from ._hed_utils import accuracies
    from .._image_utils import _get_image_chips, _get_transformed_predictions, _draw_predictions, _exclude_detection
    from .._video_utils import VideoUtils
    import inspect
    HAS_FASTAI = True
    
except Exception as e:
    HAS_FASTAI = False

try:
    import cv2
except Exception:
    HAS_OPENCV = False

try:
    import arcpy
except Exception:
    HAS_ARCPY = False    

class ModelExtension(ArcGISModel):
    """
    Creates a ModelExtension object, to train the model for object detection, semantic segmentation, and edge detection.

    =====================   ============================================================
    **Argument**            **Description**
    ---------------------   ------------------------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            ``prepare_data`` function.
    ---------------------   ------------------------------------------------------------
    model_conf              A class definition contains the following methods:

                                * ``get_model(self, data, backbone=None)``: for model definition,
                                
                                * ``on_batch_begin(self, learn, model_input_batch, model_target_batch)``: for 
                                  feeding input to the model during training, 

                                * ``transform_input(self, xb)``: for feeding input to the model during
                                  inferencing/validation,

                                * ``transform_input_multispectral(self, xb)``: for feeding input to the
                                  model during inferencing/validation in case of multispectral data,

                                * ``loss(self, model_output, *model_target)``: to return loss value of the model, and 

                                * ``post_process(self, pred, nms_overlap, thres, chip_size, device)``: to post-process
                                  the output of the object-detection model.

                                * ``post_process(self, pred, thres)``: to post-process the output of the segmentation model.
                                  
    ---------------------   ------------------------------------------------------------
    backbone                Optional function. If custom model requires any backbone.
    ---------------------   ------------------------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ============================================================

    :return: ``ModelExtension`` Object
    """

    def __init__(self, data, model_conf, backbone=None, pretrained_path=None, **kwargs):

        super().__init__(data, backbone, **kwargs)
        self.model_conf = model_conf()
        self.model_conf_class  = model_conf
        self._backend = 'pytorch'
        self._kwargs = kwargs
        model = self.model_conf.get_model(data, backbone, **kwargs)
        if self._is_multispectral:
            model = _change_tail(model, data)
        if not _isnotebook() and os.name=='posix':
            _set_ddp_multigpu(self)
            if self._multigpu_training:
                self.learn = Learner(data, model, loss_func=self.model_conf.loss).to_distributed(self._rank_distributed)
            else:
                self.learn = Learner(data, model, loss_func=self.model_conf.loss)
        else:
            self.learn = Learner(data, model, loss_func=self.model_conf.loss)
        self.learn.callbacks.append(self._train_callback(self.learn, self.model_conf.on_batch_begin))
        if self._data.dataset_type == 'Classified_Tiles':
            if getattr(self, "_is_edge_detection", False):
                from ._hed_utils import accuracy, f1_score
                self.learn.metrics = [accuracy, f1_score]
            else:
                from ._psp_utils import accuracy
                self.learn.metrics = [accuracy]
            self._code = image_classifier_prf
        else:
            self._code = code
        self._arcgis_init_callback() # make first conv weights learnable
        self._bind_dataset_methods()
        if pretrained_path is not None:
            self.load(pretrained_path)

    if HAS_FASTAI:
        class _train_callback(LearnerCallback):

            def __init__(self, learn, on_batch_begin_fn):
                super().__init__(learn)
                self.on_batch_begin_fn = on_batch_begin_fn

            def on_batch_begin(self, last_input, last_target, train, **kwargs):

                last_input, last_target = self.on_batch_begin_fn(self.learn, last_input, last_target)

                return {'last_input':last_input, 'last_target':last_target}

    def _analyze_pred(self, pred, thresh=0.5, nms_overlap=0.1, ret_scores=True, device=None):
        return self.model_conf.post_process(pred, nms_overlap, thresh, self.learn.data.chip_size, device)
       
    def _get_emd_params(self, save_inference_file):
        import random
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        if self._data.dataset_type == 'Classified_Tiles':
            _emd_template["ModelType"] = "ImageClassification"
            if save_inference_file:
                _emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
            else:
                _emd_template["InferenceFunction"] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageClassifier.py"
            _emd_template['IsEdgeDetection'] = getattr(self, "_is_edge_detection", False)
        else:
            _emd_template["ModelType"] = "ObjectDetection"
            if save_inference_file:
                _emd_template["InferenceFunction"] = "ArcGISObjectDetector.py"
            else:
                _emd_template["InferenceFunction"] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISObjectDetector.py"
        _emd_template["ModelConfiguration"] = "_model_extension_inferencing"
        _emd_template["ExtractBands"] = [0, 1, 2]
        _emd_template['Classes'] = []
        _emd_template['ModelConfigurationFile'] = "ModelConfiguration.py"
        _emd_template['ModelFileConfigurationClass'] = type(self.model_conf).__name__
        _emd_template['DatasetType'] = self._data.dataset_type
        _emd_template['Kwargs'] = self._kwargs

        class_data = {}
        for i, class_name in enumerate(self._data.classes[1:]):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            _emd_template['Classes'].append(class_data.copy())

        return _emd_template

    @property
    def _is_model_extension(self):
        return True

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a ``ModelExtension`` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from ``prepare_data`` function or None for
                                inferencing.

        =====================   ===========================================

        :returns: `ModelExtension` Object
        """

        emd_path = _get_emd_path(emd_path)

        with open(emd_path) as f:
            emd = json.load(f)
            
        model_file = Path(emd['ModelFile'])
        
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file
        
        modelconf = Path(emd['ModelConfigurationFile'])

        if not modelconf.is_absolute():
            modelconf = emd_path.parent / modelconf

        modelconfclass = emd['ModelFileConfigurationClass']

        sys.path.append(os.path.dirname(modelconf))
        model_configuration = getattr(importlib.import_module('{}'.format(modelconf.name[0:-3])), modelconfclass)

        backbone = emd['ModelParameters']['backbone']
        dataset_type = emd.get('DatasetType', 'PASCAL_VOC_rectangles')
        chip_size = emd["ImageWidth"]
        resize_to = emd.get('resize_to', None)
        kwargs = emd.get('Kwargs', {})
        if isinstance(resize_to, list):
            resize_to = (resize_to[0], resize_to[1])

        try:
            class_mapping = {i['Value'] : i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value'] : i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue'] : i['ClassName'] for i in emd['Classes']} 
            color_mapping = {i['ClassValue'] : i['Color'] for i in emd['Classes']}                

        data_passed = True
        if data is None:           

            data_passed = False
            if dataset_type == 'PASCAL_VOC_rectangles':
                train_tfms = []
                val_tfms = []
                ds_tfms = (train_tfms, val_tfms)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", UserWarning)
                    sd = ImageList([], path=emd_path.parent.parent).split_by_idx([])
                    data = sd.label_const(0, label_cls=ObjectDetectionCategoryList, classes=list(class_mapping.values())).transform(ds_tfms).databunch().normalize(imagenet_stats)
                # Add 1 for background class
                data.c += 1
            else:
                data = _EmptyData(path=emd_path.parent.parent, loss_func=None, c=len(class_mapping) + 1, chip_size=emd['ImageHeight'])
            
            data.chip_size = chip_size
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data.classes = ['background'] + list(class_mapping.values())
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data = get_multispectral_data_params_from_emd(data, emd)
            data.dataset_type = dataset_type

        data.resize_to = resize_to
        mextnsn = cls(data, model_configuration, backbone, pretrained_path=str(model_file), **kwargs)

        if not data_passed and dataset_type == 'PASCAL_VOC_rectangles':
            mextnsn.learn.data.single_ds.classes = mextnsn._data.classes
            mextnsn.learn.data.single_ds.y.classes = mextnsn._data.classes

        return mextnsn

    @property
    def _model_metrics(self):
        if self._data.dataset_type == 'Classified_Tiles':
            return {'accuracy': '{0:1.4e}'.format(self._get_model_metrics())}
        else:
            return {'average_precision_score': self.average_precision_score(show_progress=False)}

    def _get_model_metrics(self, **kwargs):
        checkpoint = kwargs.get('checkpoint', True)
        if not hasattr(self.learn, 'recorder'):
            return 0.0
        model_accuracy = self.learn.recorder.metrics[-1][0]
        if checkpoint:
            model_accuracy = np.max(self.learn.recorder.metrics)
        return float(model_accuracy)

    def _get_y(self, bbox, clas):
        try:
            bbox = bbox.view(-1, 4)
        except Exception:
            bbox = torch.zeros(size=[0, 4])
        bb_keep = ((bbox[:,2]-bbox[:,0])>0).nonzero()[:,0]
        return bbox[bb_keep],clas[bb_keep]

    def _intersect(self,box_a, box_b):
        max_xy = torch.min(box_a[:, None, 2:], box_b[None, :, 2:])
        min_xy = torch.max(box_a[:, None, :2], box_b[None, :, :2])
        inter = torch.clamp((max_xy - min_xy), min=0)
        return inter[:, :, 0] * inter[:, :, 1]

    def _box_sz(self, b):
        return (b[:, 2]-b[:, 0]) * (b[:, 3]-b[:, 1])

    def _jaccard(self, box_a, box_b):
        inter = self._intersect(box_a, box_b)
        union = self._box_sz(box_a).unsqueeze(1) + self._box_sz(box_b).unsqueeze(0) - inter
        return inter / union

    def _bind_dataset_methods(self):

        if self._data.dataset_type == 'Classified_Tiles':
            if getattr(self, "_is_edge_detection", False):
                self.show_results = self._show_results_edge_detection
                self.compute_precision_recall = self._edge_detection_accuracies
            else:
                self.show_results = self._show_results_segmentation
        else:
            if self._is_multispectral:
                self.show_results = self._show_results_multispectral
            else:
                self.show_results = self._show_results_object_detection
            self.average_precision_score = self._average_precision_score
            self.predict = self._predict
            self.predict_video = self._predict_video

    def _show_results_object_detection(self, rows=5, thresh=0.5, nms_overlap=0.1):

        """
        Displays the results of a trained model on a part of the validation set.
        """
        self._check_requisites()
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)
        self._show_results_modified(rows=rows, thresh=thresh, nms_overlap=nms_overlap, model=self)

    def _show_results_segmentation(self, rows=5, thresh=0.5, **kwargs):

        """
        Displays the results of a trained model on a part of the validation set.
        """
        self._check_requisites()
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)

        self._show_results_modified(rows=rows, thresh=thresh, model=self, **kwargs)

    def _show_results_edge_detection(self, rows=5, thresh=0.5, thinning=True,**kwargs):

        """
        Displays the results of a trained model on a part of the validation set.
        """
        self._check_requisites()
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)

        self._show_results_modified(rows=rows, thresh=thresh, model=self, thinning=thinning, **kwargs)

    def _show_results_multispectral(self, rows=5, thresh=0.3, nms_overlap=0.1, alpha=1, **kwargs):
        return_fig = kwargs.get("return_fig", False)
        ax = show_results_multispectral(
            self,
            nrows=rows,
            thresh=thresh,
            nms_overlap=nms_overlap,
            alpha=alpha,
            **kwargs
        )
        if return_fig:
            fig,axs=ax
            return fig

    def _show_results_modified(self, rows=5, **kwargs):

        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)

        ds_type = DatasetType.Valid
        n_items = rows ** 2 if self.learn.data.train_ds.x._square_show_res else rows
        if self.learn.dl(ds_type).batch_size < n_items: n_items = self.learn.dl(ds_type).batch_size
        ds = self.learn.dl(ds_type).dataset
        xb,yb = self.learn.data.one_batch(ds_type, detach=False, denorm=False)
        self.learn.model.eval()
        transform_kwargs, kwargs = split_kwargs_by_func(kwargs, self.model_conf.transform_input)
        try:
            preds = self.learn.model(self.model_conf.transform_input(xb, **transform_kwargs))
        except Exception as e:

            if getattr(self, "_is_fasterrcnn", False):
                preds = []
                for _ in range(xb.shape[0]):
                    res={}
                    res['boxes'] = torch.empty(0,4)
                    res['scores'] = torch.tensor([])
                    res['labels'] = torch.tensor([])
                    preds.append(res)
            else:
                raise e

        x,y = to_cpu(xb),to_cpu(yb)
        norm = getattr(self.learn.data,'norm',False)
        if norm:
            x = self.learn.data.denorm(x)
            if norm.keywords.get('do_y',False):
                y     = self.learn.data.denorm(y, do_x=True)
                preds = self.learn.data.denorm(preds, do_x=True)
        analyze_kwargs,kwargs = split_kwargs_by_func(kwargs, ds.y.analyze_pred)
        preds = ds.y.analyze_pred(preds,**analyze_kwargs)
        xs = [ds.x.reconstruct(grab_idx(x, i)) for i in range(n_items)]
        if has_arg(ds.y.reconstruct, 'x'):
            ys = [ds.y.reconstruct(grab_idx(y, i), x=x) for i,x in enumerate(xs)]
            zs = [ds.y.reconstruct(z, x=x) for z,x in zip(preds,xs)]
        else:
            ys = [ds.y.reconstruct(grab_idx(y, i)) for i in range(n_items)]
            zs = [ds.y.reconstruct(z) for z in preds]
        ds.x.show_xyzs(xs, ys, zs, **kwargs)

    def _predict_learn_modified(self, item, **kwargs):
        "Return predicted class, label and probabilities for `item`."
        batch = self.learn.data.one_item(item)
        transform_kwargs, kwargs = split_kwargs_by_func(kwargs, self.model_conf.transform_input)
        self.learn.model.eval()
        try:
            pred = self.learn.model(self.model_conf.transform_input(batch[0], **transform_kwargs))
        except Exception as e:

            if getattr(self, "_is_fasterrcnn", False):
                pred = []
                for _ in range(batch[0].shape[0]):
                    res={}
                    res['boxes'] = torch.empty(0,4)
                    res['scores'] = torch.tensor([])
                    res['labels'] = torch.tensor([])
                    pred.append(res)
            else:
                raise e
        ds = self.learn.data.single_ds
        analyze_kwargs,kwargs = split_kwargs_by_func(kwargs, ds.y.analyze_pred)
        pred = ds.y.analyze_pred(pred, **analyze_kwargs)
        x = batch[0]
        norm = getattr(self.learn.data,'norm',False)
        if norm:
            x = self.learn.data.denorm(x)
        x = ds.x.reconstruct(grab_idx(x, 0))
        y = ds.y.reconstruct(pred[0], x) if has_arg(ds.y.reconstruct, 'x') else ds.y.reconstruct(pred[0])
        return y

    def _average_precision_score(self, detect_thresh=0.2, iou_thresh=0.1, mean=False, show_progress=True):

        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth labels, above
                                which a predicted bounding box will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
        =====================   ===========================================

        :returns: `dict` if mean is False otherwise `float`
        """
        self._check_requisites()

        aps = compute_class_AP(self,
                               self._data.valid_dl, 
                               self._data.c - 1, 
                               show_progress, 
                               iou_thresh=iou_thresh, 
                               detect_thresh=detect_thresh, 
                               thresh=detect_thresh, 
                               nms_overlap=iou_thresh)
        if mean:
            import statistics
            return statistics.mean(aps)
        else:
            return dict(zip(self._data.classes[1:], aps))

    def _edge_detection_accuracies(self, thresh=0.5, buffer=3, show_progress=True):
        """
        Computes precision, recall and f1 score on validation set.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered edge pixel.
        ---------------------   -------------------------------------------
        buffer                  Optional int. pixels in neighborhood to
                                consider true detection.
        =====================   ===========================================

        :returns: `dict` 
        """
        self._check_requisites()
        acc = accuracies(self, self._data.valid_dl, detect_thresh=thresh, buffer=buffer, show_progress=show_progress)
        return acc

    def _predict(
        self,
        image_path,
        threshold=0.5,
        nms_overlap=0.1,
        return_scores=False,
        visualize=False,
        resize=False
    ):

        """
        Runs prediction on an Image.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        return_scores           Optional boolean. Will return the probability
                                scores of the bounding box predictions if True.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. Displays the image with
                                predicted bounding boxes if True.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the image to the same size
                                (chip_size parameter in prepare_data) that the model was trained on,
                                before detecting objects.
                                Note that if resize_to parameter was used in prepare_data,
                                the image is resized to that size instead.

                                By default, this parameter is false and the detections are run
                                in a sliding window fashion by applying the model on cropped sections
                                of the image (of the same size as the model was trained on).
        =====================   ===========================================
        
        :returns:  Returns a tuple with predictions, labels and optionally confidence scores
                   if return_scores=True. The predicted bounding boxes are returned as a list
                   of lists containing the  xmin, ymin, width and height of each predicted object
                   in each image. The labels are returned as a list of class values and the
                   confidence scores are returned as a list of floats indicating the confidence
                   of each prediction.
        """
        if not HAS_OPENCV:
            raise Exception("This function requires opencv 4.0.1.24. Install it using pip install opencv-python==4.0.1.24")

        if isinstance(image_path, str):
            image = cv2.imread(image_path)
        else:
            image = image_path

        orig_height, orig_width, _ = image.shape
        orig_frame = image.copy()

        if resize and self._data.resize_to is None\
                and self._data.chip_size is not None:
            image = cv2.resize(image, (self._data.chip_size, self._data.chip_size))

        if self._data.resize_to is not None:
            if isinstance(self._data.resize_to, tuple):
                image = cv2.resize(image, self._data.resize_to)
            else:
                image = cv2.resize(image, (self._data.resize_to, self._data.resize_to))

        height, width, _ = image.shape

        if self._data.chip_size is not None:
            chips = _get_image_chips(image, self._data.chip_size)
        else:
            chips = [{'width': width, 'height': height, 'xmin': 0, 'ymin': 0, 'chip': image, 'predictions': []}]

        include_pad_detections = False
        if len(chips) == 1:
            include_pad_detections = True

        valid_tfms = self._data.valid_ds.tfms
        self._data.valid_ds.tfms = []

        for chip in chips:
            frame = Image(pil2tensor(PIL.Image.fromarray(cv2.cvtColor(chip['chip'], cv2.COLOR_BGR2RGB)), dtype=np.float32).div_(255))
            bbox = self._predict_learn_modified(frame, thresh=threshold, nms_overlap=nms_overlap, ret_scores=True, model=self)
            if bbox:
                scores = bbox.scores
                bboxes, lbls = bbox._compute_boxes()
                bboxes.add_(1).mul_(torch.tensor([chip['height'] / 2, chip['width'] / 2, chip['height'] / 2, chip['width'] / 2])).long()
                for index, bbox in enumerate(bboxes):
                    if lbls is not None:
                        label = lbls[index]
                    else:
                        label = 'Default'

                    data = bb2hw(bbox)
                    if include_pad_detections or not _exclude_detection((data[0], data[1], data[2], data[3]), chip['width'], chip['height']):
                        chip['predictions'].append({
                            'xmin': data[0],
                            'ymin': data[1],
                            'width': data[2],
                            'height': data[3],
                            'score': float(scores[index]),
                            'label': label
                        })

        self._data.valid_ds.tfms = valid_tfms

        predictions, labels, scores = _get_transformed_predictions(chips)

        y_ratio = orig_height/height
        x_ratio = orig_width/width

        for index, prediction in enumerate(predictions):
            prediction[0] = prediction[0]*x_ratio
            prediction[1] = prediction[1]*y_ratio
            prediction[2] = prediction[2]*x_ratio
            prediction[3] = prediction[3]*y_ratio

            # Clip xmin
            if prediction[0] < 0: 
                prediction[2] = prediction[2] + prediction[0]
                prediction[0] = 1

            # Clip width when xmax greater than original width
            if prediction[0] + prediction[2] > orig_width:
                prediction[2] = (prediction[0] + prediction[2]) - orig_width

            # Clip ymin
            if prediction[1] < 0:
                prediction[3] = prediction[3] + prediction[1]
                prediction[1] = 1

            # Clip height when ymax greater than original height
            if prediction[1] + prediction[3] > orig_height:
                prediction[3] = (prediction[1] + prediction[3]) - orig_height

            predictions[index] = [
                prediction[0],
                prediction[1],
                prediction[2],
                prediction[3]
            ]      

        if visualize:
            image = _draw_predictions(orig_frame, predictions, labels)
            import matplotlib.pyplot as plt
            plt.xticks([])
            plt.yticks([])
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            plt.imshow(PIL.Image.fromarray(image))

        if return_scores:
            return predictions, labels, scores
        else:
            return predictions, labels


    def _predict_video(
            self,
            input_video_path,
            metadata_file,
            threshold=0.5,
            nms_overlap=0.1,
            track=False,
            visualize=False,
            output_file_path=None,
            multiplex=False,
            multiplex_file_path=None,
            tracker_options={
                'assignment_iou_thrd': 0.3,
                'vanish_frames': 40,
                'detect_frames': 10
            },
            visual_options={
                'show_scores': True,
                'show_labels': True,
                'thickness': 2,
                'fontface': 0,
                'color': (255, 255, 255)
            },
            resize=False
        ):

            """
            Runs prediction on a video and appends the output VMTI predictions in the metadata file.

            =====================   ===========================================
            **Argument**            **Description**
            ---------------------   -------------------------------------------
            input_video_path        Required. Path to the video file to make the
                                    predictions on.
            ---------------------   -------------------------------------------
            metadata_file           Required. Path to the metadata csv file where
                                    the predictions will be saved in VMTI format.
            ---------------------   -------------------------------------------
            threshold               Optional float. The probability above which
                                    a detection will be considered.
            ---------------------   -------------------------------------------
            nms_overlap             Optional float. The intersection over union
                                    threshold with other predicted bounding
                                    boxes, above which the box with the highest
                                    score will be considered a true positive.
            ---------------------   -------------------------------------------
            track                   Optional bool. Set this parameter as True to
                                    enable object tracking. 
            ---------------------   -------------------------------------------
            visualize               Optional boolean. If True a video is saved
                                    with prediction results.
            ---------------------   -------------------------------------------
            output_file_path        Optional path. Path of the final video to be saved.
                                    If not supplied, video will be saved at path input_video_path
                                    appended with _prediction.
            ---------------------   -------------------------------------------
            multiplex               Optional boolean. Runs Multiplex using the VMTI detections.
            ---------------------   -------------------------------------------
            multiplex_file_path     Optional path. Path of the multiplexed video to be saved.
                                    By default a new file with _multiplex.MOV extension is saved
                                    in the same folder.
            ---------------------   -------------------------------------------
            tracking_options        Optional dictionary. Set different parameters for
                                    object tracking. assignment_iou_thrd parameter is used
                                    to assign threshold for assignment of trackers, 
                                    vanish_frames is the number of frames the object should
                                    be absent to consider it as vanished, detect_frames 
                                    is the number of frames an object should be detected
                                    to track it.
            ---------------------   -------------------------------------------
            visual_options          Optional dictionary. Set different parameters for
                                    visualization.
                                    show_scores boolean, to view scores on predictions,
                                    show_labels boolean, to view labels on predictions,
                                    thickness integer, to set the thickness level of box,
                                    fontface integer, fontface value from opencv values,
                                    color tuple (B, G, R), tuple containing values between
                                    0-255.
            ---------------------   -------------------------------------------
            resize                  Optional boolean. Resizes the video frames to the same size
                                    (chip_size parameter in prepare_data) that the model was trained on,
                                    before detecting objects.
                                    Note that if resize_to parameter was used in prepare_data,
                                    the video frames are resized to that size instead.

                                    By default, this parameter is false and the detections are run
                                    in a sliding window fashion by applying the model on cropped sections
                                    of the frame (of the same size as the model was trained on).
            =====================   ===========================================
            
            """

            VideoUtils.predict_video(
                self,
                input_video_path,
                metadata_file,
                threshold,
                nms_overlap,
                track, visualize,
                output_file_path,
                multiplex,
                multiplex_file_path,
                tracker_options,
                visual_options,
                resize
            )


"""
The following license is applicable to this file only.

This software uses some portions from the following software under its license:

Copyright (c) 2018 DeNA Co., Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal 
in the Software without restriction, including without limitation the rights 
to use, copy, modify, merge, publish, distribute, and/or sublicense 
copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software; and

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


chainercv

The MIT License

Copyright (c) 2017 Preferred Networks, Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


# import necessary modules
import numpy as np
import torch
from torch import nn, LongTensor
from collections import defaultdict
import warnings

from fastprogress.fastprogress import progress_bar
from fastai.basic_train import LearnerCallback
from ._retinanet_utils import compute_ap, _get_y, IoU_values, tlbr2cthw, cthw2tlbr


def add_conv(in_ch, out_ch, ksize=3, stride=1):
    """
    Add a conv2d / batchnorm / leaky ReLU block.
    Args:
        in_ch (int): number of input channels of the convolution layer.
        out_ch (int): number of output channels of the convolution layer.
        ksize (int): kernel size of the convolution layer.
        stride (int): stride of the convolution layer.
    Returns:
        stage (Sequential) : Sequential layers composing a convolution block.
    """
    stage = nn.Sequential()
    stage.add_module('conv', nn.Conv2d(in_ch, out_ch, kernel_size=ksize, stride=stride, 
                                        padding=ksize // 2, bias=False))
    stage.add_module('batch_norm', nn.BatchNorm2d(out_ch))
    stage.add_module('leaky', nn.LeakyReLU(negative_slope=0.1, inplace=True))
    return stage


class resblock(nn.Module):
    """
    Sequential residual blocks each of which consists of \
    two convolution layers.
    Args:
        ch (int): number of input and output channels.
        nblocks (int): number of residual blocks.
        shortcut (bool): if True, residual tensor addition is enabled.
    """
    def __init__(self, ch, nblocks=1, shortcut=True):

        super().__init__()
        self.shortcut = shortcut
        self.module_list = nn.ModuleList()
        for i in range(nblocks):
            resblock_one = nn.ModuleList()
            resblock_one.append(add_conv(ch, ch//2, 1, 1))
            resblock_one.append(add_conv(ch//2, ch, 3, 1))
            self.module_list.append(resblock_one)

    def forward(self, x):
        for module in self.module_list:
            h = x
            for res in module:
                h = res(h)
            x = x + h if self.shortcut else h
        return x

def create_yolov3_modules(config_model, ignore_thre):
    """
    Build yolov3 layer modules.
    Args:
        config_model (dict): model configuration.
            See YOLOLayer class for details.
        ignore_thre (float): used in YOLOLayer.
    Returns:
        mlist (ModuleList): YOLOv3 module list.
    """
    # Fetch the number of bands (or channels) in the image
    in_ch = config_model["N_BANDS"]

    # DarkNet53
    mlist = nn.ModuleList()
    mlist.append(add_conv(in_ch=in_ch, out_ch=32, ksize=3, stride=1))               #0
    mlist.append(add_conv(in_ch=32, out_ch=64, ksize=3, stride=2))                  #1 
    mlist.append(resblock(ch=64))                                                   #2
    mlist.append(add_conv(in_ch=64, out_ch=128, ksize=3, stride=2))                 #3
    mlist.append(resblock(ch=128, nblocks=2))                                       #4
    mlist.append(add_conv(in_ch=128, out_ch=256, ksize=3, stride=2))                #5 
    mlist.append(resblock(ch=256, nblocks=8))    # shortcut 1 from here             #6 - shortcut 1
    mlist.append(add_conv(in_ch=256, out_ch=512, ksize=3, stride=2))                #7 
    mlist.append(resblock(ch=512, nblocks=8))    # shortcut 2 from here             #8 - shortcut 2
    mlist.append(add_conv(in_ch=512, out_ch=1024, ksize=3, stride=2))               #9
    mlist.append(resblock(ch=1024, nblocks=4))                                      #10

    # YOLOv3
    mlist.append(resblock(ch=1024, nblocks=2, shortcut=False))                      #11
    mlist.append(add_conv(in_ch=1024, out_ch=512, ksize=1, stride=1))               #12
    # 1st yolo branch
    mlist.append(add_conv(in_ch=512, out_ch=1024, ksize=3, stride=1))               #13
    mlist.append(
         YOLOLayer(config_model, layer_no=0, in_ch=1024, ignore_thre=ignore_thre))  #14 - yolo

    mlist.append(add_conv(in_ch=512, out_ch=256, ksize=1, stride=1))                #15
    mlist.append(nn.Upsample(scale_factor=2, mode='nearest'))                       #16 - shortcut 2 concats
    mlist.append(add_conv(in_ch=768, out_ch=256, ksize=1, stride=1))                #17
    mlist.append(add_conv(in_ch=256, out_ch=512, ksize=3, stride=1))                #18
    mlist.append(resblock(ch=512, nblocks=1, shortcut=False))                       #19
    mlist.append(add_conv(in_ch=512, out_ch=256, ksize=1, stride=1))                #20
    # 2nd yolo branch
    mlist.append(add_conv(in_ch=256, out_ch=512, ksize=3, stride=1))                #21
    mlist.append(
        YOLOLayer(config_model, layer_no=1, in_ch=512, ignore_thre=ignore_thre))    #22 - yolo

    mlist.append(add_conv(in_ch=256, out_ch=128, ksize=1, stride=1))                #23
    mlist.append(nn.Upsample(scale_factor=2, mode='nearest'))                       #24 - shortcut 1 concats
    mlist.append(add_conv(in_ch=384, out_ch=128, ksize=1, stride=1))                #25
    mlist.append(add_conv(in_ch=128, out_ch=256, ksize=3, stride=1))                #26
    mlist.append(resblock(ch=256, nblocks=2, shortcut=False))                       #27
    mlist.append(
         YOLOLayer(config_model, layer_no=2, in_ch=256, ignore_thre=ignore_thre))   #28 - yolo

    return mlist


class YOLOv3_Model(nn.Module):
    """
    YOLOv3 model module. The module list is defined by create_yolov3_modules function. \
    The network returns loss values from three YOLO layers during training \
    and detection results during test.
    """
    def __init__(self, config_model, ignore_thre=0.7):
        """
        Initialization of YOLOv3 class.
        Args:
            config_model (dict): used in YOLOLayer.
            ignore_thre (float): used in YOLOLayer.
        """
        super().__init__()
        self.module_list = create_yolov3_modules(config_model, ignore_thre)
        
    def forward(self, x, targets=None):
        """
        Forward path of YOLOv3.
        Args:
            x (torch.Tensor) : input data whose shape is :math:`(N, C, H, W)`, \
                where N, C are batchsize and num. of channels.
            targets (torch.Tensor) : label array whose shape is :math:`(N, 50, 5)`

        Returns:
            training:
                output (torch.Tensor): loss tensor for backpropagation.
            test:
                output (torch.Tensor): concatenated detection results.
        """

        train = targets is not None
        output = []
        self.loss_dict = defaultdict(float)
        route_layers = []
        for i, module in enumerate(self.module_list):
            # yolo layers
            if i in [14, 22, 28]:
                if train:
                    x, *loss_dict = module(x, targets)
                    for name, loss in zip(['xy', 'wh', 'conf', 'cls', 'l2'] , loss_dict):
                        self.loss_dict[name] += loss
                else:
                    x = module(x)
                output.append(x)
            else:
                x = module(x)

            # route layers
            if i in [6, 8, 12, 20]:
                route_layers.append(x)
            if i == 14:
                x = route_layers[2]
            if i == 22:  # yolo 2nd
                x = route_layers[3]
            if i == 16:
                x = torch.cat((x, route_layers[1]), 1)
            if i == 24:
                x = torch.cat((x, route_layers[0]), 1)
 

        if train:
            return sum(output)
        else:
            return torch.cat(output, 1)


class YOLOLayer(nn.Module):
    """
    detection layer corresponding to yolo_layer.c of darknet
    """
    def __init__(self, config_model, layer_no, in_ch, ignore_thre=0.7):
        """
        Args:
            config_model (dict) : model configuration.
                ANCHORS (list of tuples) :
                ANCH_MASK:  (list of int list): index indicating the anchors to be
                    used in YOLO layers. One of the mask group is picked from the list.
                N_CLASSES (int): number of classes
            layer_no (int): YOLO layer number - one from (0, 1, 2).
            in_ch (int): number of input channels.
            ignore_thre (float): threshold of IoU above which objectness training is ignored.
        """

        super(YOLOLayer, self).__init__()
        strides = [32, 16, 8] # fixed
        self.anchors = config_model['ANCHORS']
        self.anch_mask = config_model['ANCH_MASK'][layer_no]
        self.n_anchors = len(self.anch_mask)
        self.n_classes = config_model['N_CLASSES']
        self.ignore_thre = ignore_thre
        self.l2_loss = nn.MSELoss(reduction='sum')
        self.bce_loss = nn.BCELoss(reduction='sum')
        self.stride = strides[layer_no]
        self.all_anchors_grid = [(w / self.stride, h / self.stride)
                                 for w, h in self.anchors]
        self.masked_anchors = [self.all_anchors_grid[i]
                               for i in self.anch_mask]
        self.ref_anchors = np.zeros((len(self.all_anchors_grid), 4))
        self.ref_anchors[:, 2:] = np.array(self.all_anchors_grid)
        self.ref_anchors = torch.FloatTensor(self.ref_anchors)
        self.conv = nn.Conv2d(in_channels=in_ch,
                              out_channels=self.n_anchors * (self.n_classes + 5),
                              kernel_size=1, stride=1, padding=0)

    def forward(self, xin, labels=None):
        """
        In this
        Args:
            xin (torch.Tensor): input feature map whose size is :math:`(N, C, H, W)`, \
                where N, C, H, W denote batchsize, channel width, height, width respectively.
            labels (torch.Tensor): label data whose size is :math:`(N, K, 5)`. \
                N: batchsize and
                K: number of labels (max labels in an image (padded)).
                Each label consists of [class, xc, yc, w, h]:
                    class (float): class index.
                    xc, yc (float) : center of bbox whose values range from 0 to 1.
                    w, h (float) : size of bbox whose values range from 0 to 1.
        Returns:
            loss (torch.Tensor): total loss - the target of backprop.
            loss_xy (torch.Tensor): x, y loss - calculated by binary cross entropy (BCE) \
                with boxsize-dependent weights.
            loss_wh (torch.Tensor): w, h loss - calculated by l2 without size averaging and \
                with boxsize-dependent weights.
            loss_obj (torch.Tensor): objectness loss - calculated by BCE.
            loss_cls (torch.Tensor): classification loss - calculated by BCE for each class.
            loss_l2 (torch.Tensor): total l2 loss - only for logging.
        """

        output = self.conv(xin)

        batchsize = output.shape[0]
        fsize = output.shape[2]
        n_ch = 5 + self.n_classes

        dtype = torch.cuda.FloatTensor if xin.is_cuda else torch.FloatTensor

        output = output.view(batchsize, self.n_anchors, n_ch, fsize, fsize)
        output = output.permute(0, 1, 3, 4, 2)

        # logistic activation for xy, obj, cls
        output[..., np.r_[:2, 4:n_ch]] = torch.sigmoid(
            output[..., np.r_[:2, 4:n_ch]])

        # calculate pred - xywh obj cls
        x_shift = dtype(np.broadcast_to(
            np.arange(fsize, dtype=np.float32), output.shape[:4]))
        y_shift = dtype(np.broadcast_to(
            np.arange(fsize, dtype=np.float32).reshape(fsize, 1), output.shape[:4]))

        masked_anchors = np.array(self.masked_anchors)

        w_anchors = dtype(np.broadcast_to(np.reshape(
            masked_anchors[:, 0], (1, self.n_anchors, 1, 1)), output.shape[:4]))
        h_anchors = dtype(np.broadcast_to(np.reshape(
            masked_anchors[:, 1], (1, self.n_anchors, 1, 1)), output.shape[:4]))

        pred = output.clone()
        pred[..., 0] += x_shift
        pred[..., 1] += y_shift
        pred[..., 2] = torch.exp(pred[..., 2]) * w_anchors
        pred[..., 3] = torch.exp(pred[..., 3]) * h_anchors

        # return the predictions when not training
        if labels is None:
            pred[..., :4] *= self.stride # Scale bbox coordinates to image size
            return pred.view(batchsize, -1, n_ch).data

        pred = pred[..., :4].data

        # target assignment
        tgt_mask = torch.zeros(batchsize, self.n_anchors,
                               fsize, fsize, 4 + self.n_classes).type(dtype)
        obj_mask = torch.ones(batchsize, self.n_anchors,
                              fsize, fsize).type(dtype)
        tgt_scale = torch.zeros(batchsize, self.n_anchors,
                                fsize, fsize, 2).type(dtype)

        target = torch.zeros(batchsize, self.n_anchors,
                             fsize, fsize, n_ch).type(dtype)

        labels = labels.cpu().data
        # Rearrange the labels (dim=1) so that ground truths are ordered before paddings ([0,0,0,0,0]) 
        labels = labels.flip(1)
        # If there are no bboxes in the batch, create a zeros tensor with consistent shape
        if labels.nelement() == 0: 
            labels = torch.zeros(batchsize, 1, 5)

        nlabel = (labels.sum(dim=2) > 0).sum(dim=1)  # number of objects

        # Convert to values normalized wrt grid cell size (or stride)
        truth_x_all = labels[:, :, 1] * fsize
        truth_y_all = labels[:, :, 2] * fsize
        truth_w_all = labels[:, :, 3] * fsize
        truth_h_all = labels[:, :, 4] * fsize

        # Find the grid loc of each object
        truth_i_all = truth_x_all.to(torch.int16).numpy()
        truth_j_all = truth_y_all.to(torch.int16).numpy()

        for b in range(batchsize):
            n = int(nlabel[b])  # number of objects in the image
            if n == 0:
                continue

            truth_box = dtype(np.zeros((n, 4)))
            truth_box[:n, 2] = truth_w_all[b, :n] # w
            truth_box[:n, 3] = truth_h_all[b, :n] # h
            truth_i = truth_i_all[b, :n] # i loc in grid
            truth_j = truth_j_all[b, :n] # j loc in grid

            # calculate iou between truth and reference anchors
            anchor_ious_all = bboxes_iou(truth_box.cpu(), self.ref_anchors) #shape: (n, 9) - n: #bboxes in the image, 9: #anchors
            best_n_all = np.argmax(anchor_ious_all, axis=1) #tensor shape: (n) - indices of best matched anchor box (0-8) for each bbox
            best_n = best_n_all % 3 #tensor shape: (n) - indices from the anchor mask, values in (0, 1, 2)
            # Only select the anchors if they are in the anch_mask for this layer, values in [(0,1,2), (3,4,5), (6,7,8)]
            best_n_mask = ((best_n_all == self.anch_mask[0]) | (
                best_n_all == self.anch_mask[1]) | (best_n_all == self.anch_mask[2])) # shape: (n) values in (0,1) - either matches one of the designated anchors or not

            truth_box[:n, 0] = truth_x_all[b, :n]
            truth_box[:n, 1] = truth_y_all[b, :n]

            # calculate iou between predictions and ground truth boxes
            pred_ious = bboxes_iou(pred[b].view(-1, 4), truth_box, xyxy=False) #shape: (ch*fsize*fsize, n)
            pred_best_iou, _ = pred_ious.max(dim=1) # Find the max ious values for each pred, shape: (ch*fsize*fsize)
            pred_best_iou = (pred_best_iou > self.ignore_thre) # Create a mask using the thresh
            pred_best_iou = pred_best_iou.view(pred[b].shape[:3]) # shape: (ch, fsize, fsize)
            # set mask to zero (ignore) if pred matches truth
            obj_mask[b] = 1 - pred_best_iou.int() #

            # If none of the ground truth box matches the anchors for this layer, continue with next image
            if sum(best_n_mask) == 0:
                continue
            
            # For every ground truth box
            for ti in range(best_n.shape[0]):
                if best_n_mask[ti] == 1: # if truth box matches one of the designated anchor
                    i, j = truth_i[ti], truth_j[ti]
                    a = best_n[ti] # Index of matched anchor box, value in (0,1,2)
                    obj_mask[b, a, j, i] = 1
                    tgt_mask[b, a, j, i, :] = 1
                    target[b, a, j, i, 0] = truth_x_all[b, ti] - \
                        truth_x_all[b, ti].to(torch.int16).to(torch.float)
                    target[b, a, j, i, 1] = truth_y_all[b, ti] - \
                        truth_y_all[b, ti].to(torch.int16).to(torch.float)
                    target[b, a, j, i, 2] = torch.log(
                        truth_w_all[b, ti] / torch.Tensor(self.masked_anchors)[best_n[ti], 0] + 1e-16)
                    target[b, a, j, i, 3] = torch.log(
                        truth_h_all[b, ti] / torch.Tensor(self.masked_anchors)[best_n[ti], 1] + 1e-16)
                    target[b, a, j, i, 4] = 1
                    target[b, a, j, i, 4 + labels[b, ti, 0].to(torch.int16).numpy()] = 1
                    tgt_scale[b, a, j, i, :] = torch.sqrt(
                        2 - truth_w_all[b, ti] * truth_h_all[b, ti] / fsize / fsize)
                    
        # loss calculation

        output[..., 4] *= obj_mask
        output[..., np.r_[0:4, 5:n_ch]] *= tgt_mask
        output[..., 2:4] *= tgt_scale

        target[..., 4] *= obj_mask
        target[..., np.r_[0:4, 5:n_ch]] *= tgt_mask
        target[..., 2:4] *= tgt_scale

        bceloss = nn.BCELoss(weight=tgt_scale*tgt_scale, reduction='sum')  # weighted BCEloss
        loss_xy = bceloss(output[..., :2], target[..., :2])
        loss_wh = self.l2_loss(output[..., 2:4], target[..., 2:4]) / 2
        loss_obj = self.bce_loss(output[..., 4], target[..., 4])
        loss_cls = self.bce_loss(output[..., 5:], target[..., 5:])
        loss_l2 = self.l2_loss(output, target)

        loss = (loss_xy + loss_wh + loss_obj + loss_cls).to(torch.float)

        return loss, loss_xy, loss_wh, loss_obj, loss_cls, loss_l2


class YOLOv3_Loss(nn.Module):

    def __init__(self):
        super().__init__()

    def forward(self, output, bbox_tgts, clas_tgts):
        #YOLOv3 model itself outputs loss when training
        return output


def bboxes_iou(bboxes_a, bboxes_b, xyxy=True):
    """Calculate the Intersection of Unions (IoUs) between bounding boxes.
    IoU is calculated as a ratio of area of the intersection
    and area of the union.

    Args:
        bbox_a (array): An array whose shape is :math:`(N, 4)`.
            :math:`N` is the number of bounding boxes.
            The dtype should be :obj:`numpy.float32`.
        bbox_b (array): An array similar to :obj:`bbox_a`,
            whose shape is :math:`(K, 4)`.
            The dtype should be :obj:`numpy.float32`.
    Returns:
        array:
        An array whose shape is :math:`(N, K)`. \
        An element at index :math:`(n, k)` contains IoUs between \
        :math:`n` th bounding box in :obj:`bbox_a` and :math:`k` th bounding \
        box in :obj:`bbox_b`.

    from: https://github.com/chainer/chainercv
    """
    if bboxes_a.shape[1] != 4 or bboxes_b.shape[1] != 4:
        raise IndexError

    # top left
    if xyxy:
        tl = torch.max(bboxes_a[:, None, :2], bboxes_b[:, :2])
        # bottom right
        br = torch.min(bboxes_a[:, None, 2:], bboxes_b[:, 2:])
        area_a = torch.prod(bboxes_a[:, 2:] - bboxes_a[:, :2], 1)
        area_b = torch.prod(bboxes_b[:, 2:] - bboxes_b[:, :2], 1)
    else:
        tl = torch.max((bboxes_a[:, None, :2] - bboxes_a[:, None, 2:] / 2),
                        (bboxes_b[:, :2] - bboxes_b[:, 2:] / 2))
        # bottom right
        br = torch.min((bboxes_a[:, None, :2] + bboxes_a[:, None, 2:] / 2),
                        (bboxes_b[:, :2] + bboxes_b[:, 2:] / 2))

        area_a = torch.prod(bboxes_a[:, 2:], 1)
        area_b = torch.prod(bboxes_b[:, 2:], 1)
    en = (tl < br).type(tl.type()).prod(dim=2)
    area_i = torch.prod(br - tl, 2) * en  # * ((tl < br).all())
    return area_i / (area_a[:, None] + area_b - area_i)

def nms(bbox, thresh, score=None, limit=None):
    """Suppress bounding boxes according to their IoUs and confidence scores.
    Args:
        bbox (array): Bounding boxes to be transformed. The shape is
            :math:`(R, 4)`. :math:`R` is the number of bounding boxes.
        thresh (float): Threshold of IoUs.
        score (array): An array of confidences whose shape is :math:`(R,)`.
        limit (int): The upper bound of the number of the output bounding
            boxes. If it is not specified, this method selects as many
            bounding boxes as possible.
    Returns:
        array:
        An array with indices of bounding boxes that are selected. \
        They are sorted by the scores of bounding boxes in descending \
        order. \
        The shape of this array is :math:`(K,)` and its dtype is\
        :obj:`numpy.int32`. Note that :math:`K \\leq R`.

    from: https://github.com/chainer/chainercv
    """

    if len(bbox) == 0:
        return np.zeros((0,), dtype=np.int32)

    if score is not None:
        order = score.argsort()[::-1]
        bbox = bbox[order]
    bbox_area = np.prod(bbox[:, 2:] - bbox[:, :2], axis=1)

    selec = np.zeros(bbox.shape[0], dtype=bool)
    for i, b in enumerate(bbox):
        tl = np.maximum(b[:2], bbox[selec, :2])
        br = np.minimum(b[2:], bbox[selec, 2:])
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            area = np.array(np.prod(br - tl, axis=1) * (tl < br).all(axis=1), dtype=np.float64)
            iou = area / (bbox_area[i] + bbox_area[selec] - area)
            if (iou >= thresh).any():
                continue

        selec[i] = True
        if limit is not None and np.count_nonzero(selec) >= limit:
            break

    selec = np.where(selec)[0]
    if score is not None:
        selec = order[selec]
    return selec.astype(np.int32)

def postprocess(prediction, chip_size, conf_thre=0.7, nms_thre=0.45):
    """
    Postprocess the output of YOLO model,
    perform box transformation, specify the class for each detection,
    and perform class-wise non-maximum suppression.
    Args:
        prediction (torch tensor): The shape is :math:`(N, B, 4+1+k)`.
            :math:`N` is the number of predictions,
            :math:`B` the number of boxes. The last axis consists of
            :math:`xc, yc, w, h, obj_score, cls_scores` 
            where `xc` and `yc` represent a center of a bounding box.
        chip_size (int):
            size of the images.
        conf_thre (float):
            confidence threshold ranging from 0 to 1, default 0.7
            which is defined in the config file.
        nms_thre (float):
            IoU threshold of non-max suppression ranging from 0 to 1. default 0.45

    Returns:
        output (list of torch tensor):

    """
    # Adding a dimension to handle a single image prediction
    if len(prediction.shape) == 2:
        prediction.unsqueeze_(0)

    # Convert bboxes from cthw to tlbr
    box_corner = prediction.new(prediction.shape)
    box_corner[:, :, 0] = prediction[:, :, 0] - prediction[:, :, 2] / 2
    box_corner[:, :, 1] = prediction[:, :, 1] - prediction[:, :, 3] / 2
    box_corner[:, :, 2] = prediction[:, :, 0] + prediction[:, :, 2] / 2
    box_corner[:, :, 3] = prediction[:, :, 1] + prediction[:, :, 3] / 2
    prediction[:, :, :4] = box_corner[:, :, :4]

    output = [None for _ in range(len(prediction))]
    for i, image_pred in enumerate(prediction):
        # Filter out confidence scores below threshold
        class_pred = torch.max(image_pred[:, 5:], 1)
        class_pred = class_pred[0] # Taking the class prediction scores
        conf_mask = (image_pred[:, 4] * class_pred >= conf_thre).squeeze()
        image_pred = image_pred[conf_mask]

        # If none are remaining => process next image
        if not image_pred.size(0):
            continue
        
        # Get detections with higher confidence scores than the threshold
        ind = (image_pred[:, 5:] * image_pred[:, 4][:, None] >= conf_thre).nonzero()
        
        # Detections ordered as (x1, y1, x2, y2, obj_conf, class_conf, class_pred)
        detections = torch.cat((
                image_pred[ind[:, 0], :5],
                image_pred[ind[:, 0], 5 + ind[:, 1]].unsqueeze(1),
                ind[:, 1].float().unsqueeze(1)
                ), 1)
        
        # Iterate through all predicted classes
        unique_labels = detections[:, -1].cpu().unique()
        if prediction.is_cuda:
            unique_labels = unique_labels.cuda()
        for c in unique_labels:
            # Get the detections with the particular class
            detections_class = detections[detections[:, -1] == c]
            nms_in = detections_class.cpu().numpy()
            nms_out_index = nms(
                nms_in[:, :4], thresh=nms_thre, score=nms_in[:, 4]*nms_in[:, 5])
            detections_class = detections_class[nms_out_index]
            if output[i] is None:
                output[i] = detections_class
            else:
                output[i] = torch.cat((output[i], detections_class))
    
    if output[0] is None: return None # when there is no detection
    
    bbox_scaled = torch.clamp(output[0][:,:4]/(chip_size-1) * 2 - 1, min=-1, max=1)  # Scale: -1 to 1
    bbox_pred = torch.index_select(bbox_scaled, dim=1, index=torch.tensor([1,0,3,2]).to(bbox_scaled.device)) # bbox ordered as y1,x1,y2,x2
    preds = output[0][:,-1].to(torch.long) + 1 # Add 1 to account for background class
    scores = output[0][:,4] * output[0][:,5]
    
    return bbox_pred, preds, scores


def transform_targets(last_target):
    """
    Args:
        last_target: List[Tensor(B,N,4), Tensor(B,N)] 
    Returns:
        targets (torch.tensor): shape (B, N, 5)
    """
    B, N, _ = last_target[0].shape

    lt = torch.cat((last_target[1].unsqueeze(2).type(torch.float), last_target[0].type(torch.float)), dim=2)
    lt = torch.reshape(lt, (-1,5))
    tmask = lt[:,0] != 0. # Mask to use to transform only non-zero labels
    target = torch.zeros_like(lt)

    lt[tmask, 1:] = (lt[tmask, 1:] + 1) / 2 # Scale between 0 and 1 from -1 to 1
    target[:,0] = lt[:,0]
    target[:,1] = (lt[:,2] + lt[:,4]) / 2
    target[:,2] = (lt[:,1] + lt[:,3]) / 2
    target[:,3] = lt[:,4] - lt[:,2]
    target[:,4] = lt[:,3] - lt[:,1]

    target = torch.reshape(target, (B,N,5))
    return target


class AppendLabelsCallback(LearnerCallback):
    """
    This callback is used to append labels with input
    to be passed to the YOLOv3 while training.
    """
    def __init__(self, learn):
        super().__init__(learn)
        
    def on_batch_begin(self, last_input, last_target, **kwargs):
        "Returns xb (images,labels), yb (labels) when training"
        if self.learn.predicting:
            self.learn.predicting = False
            return {'last_input':last_input, 'last_target':last_target}      
        else:
            # On training set xb as (inputs, targets) because YOLOv3 needs both
            self.learn.model.train()
            targets = transform_targets(last_target)
            return {'last_input':(last_input, targets), 'last_target':last_target}


from ._ssd_utils import kmeans, avg_iou
def generate_anchors(num_anchor, hw, limit=1000):
    """
    Function to generate anchors using k-means
    Args: 
        num_anchors (int) - number of anchors to generate
        hw - List of height width of all bounding boxes in the dataset.
        limit - max num of bounding boxes to consider for k-means
    Returns: 
        a list of anchors, shape(num_anchors, 2)
    """

    if limit > len(hw): limit = len(hw)
    import random
    idx = random.sample(range(len(hw)), limit)
    hw = np.array(hw)[idx]
    new_centroid = kmeans(hw, num_anchor)
    anchors = (np.ceil(new_centroid)).astype(int)
    anchors = anchors.tolist()
    anchors.sort(key = lambda x: x[0]*x[1])
    return update_size_zero_anchors(anchors)


def update_size_zero_anchors(anchors):
    """
    Function to replace zero size anchors [0,0] 
    by incrementally (+1) bigger anchors than the largest one.
    Args: 
        anchors, shape(N, 2), where N is number of anchors
    Returns:
        an updated list of anchors.
    """
   
    zero_anchors = sum([anchor == [0,0] for anchor in anchors])
    largest_anchor = anchors[-1]
    
    for i in range(zero_anchors):
        largest_anchor = [j+1 for j in largest_anchor]
        anchors.append(largest_anchor)
    
    return anchors[zero_anchors:]


def compute_class_AP(model, dl, n_classes, show_progress, iou_thresh=0.1, detect_thresh=0.5, num_keep=100):

    tps, clas, p_scores = [], [], []
    classes, n_gts = LongTensor(range(n_classes)),torch.zeros(n_classes).long()
    model.learn.model.eval()
    
    with torch.no_grad():
        for input,target in progress_bar(dl, display=show_progress):
            # input, shape: B, C, H, W 
            model.learn.predicting = True
            output = model.learn.pred_batch(batch=(input, target))

            for i in range(target[0].size(0)): # range batch-size
                #op - bbox preds, class preds, scores
                op = model._data.y.analyze_pred(output[i], model=model,thresh=detect_thresh, nms_overlap=iou_thresh, 
                                                ret_scores=True, device=model._device)

                # Unpad the targets
                tgt_bbox, tgt_clas = _get_y(target[0][i], target[1][i])
                
                try:
                    bbox_pred, preds, scores = op
                    if len(bbox_pred) != 0 and len(tgt_bbox) != 0:
                        
                        bbox_pred = bbox_pred.to(model._device)
                        preds = preds.to(model._device)
                        tgt_bbox = tgt_bbox.to(model._device)
                        
                        # Convert the bbox coordinates to center-height-width(cthw) before calculating Intersection Over Union
                        ious = IoU_values(tlbr2cthw(bbox_pred), tlbr2cthw(tgt_bbox))
                        max_iou, matches = ious.max(1)
                        detected = []
                    
                        for i in range(len(preds)):
                            if max_iou[i] >= iou_thresh and matches[i] not in detected and tgt_clas[matches[i]] == preds[i]:
                                detected.append(matches[i])
                                tps.append(1)
                            else: tps.append(0)
                        clas.append(preds.cpu())
                        p_scores.append(scores.cpu())
                except:
                    pass
                n_gts += ((tgt_clas.cpu()[:,None] - 1) == classes[None,:]).sum(0)             
    
    # If no true positives are found return an average precision score of 0.
    if len(tps) == 0: return [0. for cls in range(1,n_classes+1)] 

    tps, p_scores, clas = torch.tensor(tps), torch.cat(p_scores,0), torch.cat(clas,0)
    fps = 1-tps
    idx = p_scores.argsort(descending=True)
    tps, fps, clas = tps[idx], fps[idx], clas[idx]
    aps = []

    for cls in range(1,n_classes+1):
        tps_cls, fps_cls = tps[clas==cls].float().cumsum(0), fps[clas==cls].float().cumsum(0)
        if tps_cls.numel() != 0 and tps_cls[-1] != 0:
            precision = tps_cls / (tps_cls + fps_cls + 1e-8)
            recall = tps_cls / (n_gts[cls - 1] + 1e-8)
            aps.append(compute_ap(precision, recall))
        else: aps.append(0.)
    return aps


def parse_conv_block(m, weights, offset, initflag):
    """
    Initialization of conv layers with batchnorm
    Args:
        m (Sequential): sequence of layers
        weights (numpy.ndarray): pretrained weights data
        offset (int): current position in the weights file
        initflag (bool): if True, the layers are not covered by the weights file. \
            They are initialized using darknet-style initialization.
    Returns:
        offset (int): current position in the weights file
        weights (numpy.ndarray): pretrained weights data
    """
    conv_model = m[0]
    bn_model = m[1]
    param_length = m[1].bias.numel()

    # batchnorm
    for pname in ['bias', 'weight', 'running_mean', 'running_var']:
        layerparam = getattr(bn_model, pname)

        if initflag: # yolo initialization - scale to one, bias to zero
            if pname == 'weight':
                weights = np.append(weights, np.ones(param_length))
            else:
                weights = np.append(weights, np.zeros(param_length))

        param = torch.from_numpy(weights[offset:offset + param_length]).view_as(layerparam)
        layerparam.data.copy_(param)
        offset += param_length

    param_length = conv_model.weight.numel()

    # conv
    if initflag: # yolo initialization
        n, c, k, _ = conv_model.weight.shape
        scale = np.sqrt(2 / (k * k * c))
        weights = np.append(weights, scale * np.random.normal(size=param_length))

    param = torch.from_numpy(
        weights[offset:offset + param_length]).view_as(conv_model.weight)
    conv_model.weight.data.copy_(param)
    offset += param_length

    return offset, weights

def parse_yolo_block(m, weights, offset, initflag):
    """
    YOLO Layer (one conv with bias) Initialization
    Args:
        m (Sequential): sequence of layers
        weights (numpy.ndarray): pretrained weights data
        offset (int): current position in the weights file
        initflag (bool): if True, the layers are not covered by the weights file. \
            They are initialized using darknet-style initialization.
    Returns:
        offset (int): current position in the weights file
        weights (numpy.ndarray): pretrained weights data
    """
    conv_model = m._modules['conv']
    param_length = conv_model.bias.numel()

    if initflag: # yolo initialization - bias to zero
        weights = np.append(weights, np.zeros(param_length))

    param = torch.from_numpy(
        weights[offset:offset + param_length]).view_as(conv_model.bias)
    conv_model.bias.data.copy_(param)
    offset += param_length

    param_length = conv_model.weight.numel()

    if initflag: # yolo initialization
        n, c, k, _ = conv_model.weight.shape
        scale = np.sqrt(2 / (k * k * c))
        weights = np.append(weights, scale * np.random.normal(size=param_length))
 
    param = torch.from_numpy(
        weights[offset:offset + param_length]).view_as(conv_model.weight)
    conv_model.weight.data.copy_(param)
    offset += param_length

    return offset, weights

def parse_yolo_weights(model, weights_path):
    """
    Parse YOLO (darknet) pre-trained weights data onto the pytorch model
    Args:
        model : pytorch model object
        weights_path (str): path to the YOLO (darknet) pre-trained weights file
    """
    fp = open(weights_path, "rb")

    # skip the header
    header = np.fromfile(fp, dtype=np.int32, count=5) # not used
    # read weights 
    weights = np.fromfile(fp, dtype=np.float32)
    fp.close()

    offset = 0 
    initflag = False #whole yolo weights : False, darknet weights : True

    for m in model.module_list:

        if m._get_name() == 'Sequential':
            # normal conv block
            offset, weights = parse_conv_block(m, weights, offset, initflag)

        elif m._get_name() == 'resblock':
            # residual block
            for modu in m._modules['module_list']:
                for blk in modu:
                    offset, weights = parse_conv_block(blk, weights, offset, initflag)

        elif m._get_name() == 'YOLOLayer':
            # YOLO Layer (one conv with bias) Initialization
            offset, weights = parse_yolo_block(m, weights, offset, initflag)

        initflag = (offset >= len(weights)) # the end of the weights file. turn the flag on


def download_yolo_weights(weights_path):
    """ Download COCO pretrained weights for YOLOv3. """
    from arcgis.gis import GIS
    gis = GIS(set_active=False)
    item = gis.content.get('8b4600eb9a29407bbfe51491ad5bf62c')
    print(f"[INFO] Downloading COCO pretrained weights for YOLOv3 in {weights_path}...")
    filepath = item.download(weights_path)
    return

def coco_config():
    """ Function to return YOLOv3 model configurations for COCO dataset. """
    config_model= {}
    config_model['ANCHORS'] =   [[10, 13], [16, 30], [33, 23],
                                [30, 61], [62, 45], [59, 119],
                                [116, 90], [156, 198], [373, 326]]
    config_model['ANCH_MASK'] = [[6, 7, 8], [3, 4, 5], [0, 1, 2]]
    config_model['N_CLASSES'] = 80
    config_model['N_BANDS'] =   3
    return config_model

def coco_class_mapping():
    """ Create class mapping for COCO dataset. """

    # 80 COCO class indices on which YOLOv3 is pretrained
    coco_class_ids = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 20,
                    21, 22, 23, 24, 25, 27, 28, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44,
                    46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 67,
                    70, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 84, 85, 86, 87, 88, 89, 90]

    # 90 classes of COCO dataset
    coco_label_names = ('background', 
                        'person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train', 'truck',
                        'boat', 'traffic light', 'fire hydrant', 'street sign', 'stop sign',
                        'parking meter', 'bench', 'bird', 'cat', 'dog', 'horse', 'sheep', 'cow',
                        'elephant', 'bear', 'zebra', 'giraffe', 'hat', 'backpack', 'umbrella',
                        'shoe', 'eye glasses', 'handbag', 'tie', 'suitcase', 'frisbee', 'skis',
                        'snowboard', 'sports ball', 'kite', 'baseball bat', 'baseball glove',
                        'skateboard', 'surfboard', 'tennis racket', 'bottle', 'plate', 'wine glass',
                        'cup', 'fork', 'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich',
                        'orange', 'broccoli', 'carrot', 'hot dog', 'pizza', 'donut', 'cake', 'chair',
                        'couch', 'potted plant', 'bed', 'mirror', 'dining table', 'window', 'desk',
                        'toilet', 'door', 'tv', 'laptop', 'mouse', 'remote', 'keyboard', 'cell phone',
                        'microwave', 'oven', 'toaster', 'sink', 'refrigerator', 'blender', 'book',
                        'clock', 'vase', 'scissors', 'teddy bear', 'hair drier', 'toothbrush'
                        )

    coco_class_mapping = {k:v for k, v in enumerate(coco_label_names)}
    class_mapping = {k:v for k, v in coco_class_mapping.items() if k in coco_class_ids}

    return class_mapping
